
from epsilon.hotfix import require

require("twisted", "trial_assertwarns")

from axiom.store import Store
from axiom.item import Item
from axiom.attributes import integer, text, AND
from twisted.trial import unittest

from xmantissa import tdb, scrolltable

class X(Item):
    typeName = 'test_tdb_model_dummy'
    schemaVersion = 1

    number = integer()
    textNumber = text()
    phoneticDigits = text()

digits = ['zero',
          'one',
          'two',
          'three',
          'four',
          'five',
          'six',
          'seven',
          'eight',
          'nine']

def breakdown(x):
    for c in x:
        yield digits[int(c)]


class UnsortableColumn(scrolltable.AttributeColumn):
    def sortAttribute(self):
        return None


class DeprecatedNamesTest(unittest.TestCase):
    """
    Verify that deprecated names in the 'tdb' module warn appropriately.
    """

    def test_attributeColumn(self):
        """
        Verify that using the AttributeColumn name imported from 'tdb' will warn
        the user when it is instantiated.
        """
        def function():
            attrcol = tdb.AttributeColumn(X.number)
        self.assertWarns(
            DeprecationWarning,
            "tdb.AttributeColumn is deprecated.  "
            "Use scrolltable.AttributeColumn instead.",
            __file__, function)



class ModelTest(unittest.TestCase):

    def setUp(self):
        self.store = Store()
        def _():
            for x in range(107):
                X(store=self.store,
                number=x,
                textNumber=unicode(x),
                phoneticDigits=u' '.join(breakdown(str(x))))
        self.store.transact(_)

    def assertNumbersAre(self, tdm, seq):
        self.assertEquals(list(x.get('number') for x in tdm.currentPage()), seq)

    def testUniformValues(self):
        for x in self.store.query(X):
            x.number = 1

        tdm = tdb.TabularDataModel(self.store,
                X, [X.number],
                itemsPerPage=15)

        self.failUnless(tdm.hasNextPage(), 'expected there to be a next page')


    def testDeleteEverything(self):
        tdm = tdb.TabularDataModel(self.store,
                X, [X.number,
                    X.textNumber,
                    X.phoneticDigits],
                itemsPerPage=15)

        self.assertNumbersAre(tdm, range(15))
        for item in self.store.query(X):
            item.deleteFromStore()
        self.assertNumbersAre(tdm, [])

    def testOnePage(self):
        tdm = tdb.TabularDataModel(self.store,
                X, [X.number,
                    X.textNumber,
                    X.phoneticDigits],
                AND(X.number >= 17,
                    X.number < 94),
                itemsPerPage=15)

        _assertNumbersAre = lambda seq: self.assertNumbersAre(tdm, seq)

        _assertNumbersAre(range(17, 17+15))
        tdm.nextPage()
        _assertNumbersAre(range(17+15, 17+30))
        tdm.firstPage()
        _assertNumbersAre(range(17, 17+15))

    def testLeafing(self):
        tdm = tdb.TabularDataModel(self.store,
                X, [X.number,
                    X.textNumber,
                    X.phoneticDigits],
                itemsPerPage=15)

        _assertNumbersAre = lambda seq: self.assertNumbersAre(tdm, seq)

        _assertNumbersAre(range(15))
        tdm.nextPage()
        _assertNumbersAre(range(15, 30))
        tdm.prevPage()
        _assertNumbersAre(range(15))
        tdm.lastPage()
        # This next assert is kind of weak, because it is valid to have only a
        # few results on the last page, but right now the strategy is to
        # always keep the page full of results.  No matter what though, the
        # last page should contain the last value.
        self.assertEquals(list(tdm.currentPage())[-1]['number'],
                          106)
        tdm.firstPage()
        _assertNumbersAre(range(15))

    def testLeafingAndSorting(self):
        tdm = tdb.TabularDataModel(self.store,
                X, [X.number,
                    X.textNumber,
                    X.phoneticDigits],
                itemsPerPage=15)

        _assertNumbersAre = lambda seq: self.assertNumbersAre(tdm, seq)

        _assertNumbersAre(range(15))
        tdm.lastPage()
        _assertNumbersAre(range(107-15, 107))
        tdm.prevPage()
        _assertNumbersAre(range(107-30, 107-15))
        tdm.firstPage()
        _assertNumbersAre(range(15))
        tdm.resort('number', False)
        tdm.firstPage()
        _assertNumbersAre(list(reversed(range(107-15, 107))))
        tdm.lastPage()
        _assertNumbersAre(list(reversed(range(15))))

    def testPageUnderrun(self):
        tdm = tdb.TabularDataModel(self.store,
                X, [X.number,
                    X.textNumber,
                    X.phoneticDigits],
                itemsPerPage=15)
        _assertNumbersAre = lambda seq: self.assertNumbersAre(tdm, seq)

        _assertNumbersAre(range(15))
        # go to the pre-penultimate page
        for i in xrange(5):
            tdm.nextPage()
        _assertNumbersAre(range(15*5, (15*5)+15))
        tdm.nextPage()
        lastFullSet = range(15*6, (15*6)+15)
        _assertNumbersAre(lastFullSet)
        tdm.nextPage()
        lastSet = [105, 106]
        _assertNumbersAre(lastSet)

    def testItemUnderrun(self):
        tdm = tdb.TabularDataModel(self.store,
                X, [X.number,
                    X.textNumber,
                    X.phoneticDigits],
                itemsPerPage=110)

        assertFirstPage = lambda: self.assertNumbersAre(tdm, range(107))

        assertFirstPage()
        tdm.nextPage()
        assertFirstPage()
        tdm.prevPage()
        assertFirstPage()
        tdm.lastPage()
        assertFirstPage()
        tdm.firstPage()
        assertFirstPage()

    def testTwoPageNextLastEquality(self):
        tdm = tdb.TabularDataModel(self.store,
                                   X, [X.number],
                                   itemsPerPage=100)

        assertFirstPage = lambda: self.assertNumbersAre(tdm, range(100))
        assertSecondPage = lambda: self.assertNumbersAre(tdm, range(100, 107))

        assertFirstPage()
        tdm.nextPage()
        assertSecondPage()
        tdm.firstPage()
        assertFirstPage()
        tdm.lastPage()
        assertSecondPage()
    testTwoPageNextLastEquality.todo = 'is this a bug?  it seems like one to me'

    def testPagination(self):
        tdm = tdb.TabularDataModel(self.store,
                                   X, [X.number,
                                       X.textNumber,
                                       X.phoneticDigits],
                                   itemsPerPage=15)
        self.assertEquals(tdm.pageNumber, 1)
        self.assertEquals(tdm.totalItems, 107)
        self.assertEquals(tdm.totalPages, 8)

        tdm.nextPage()
        self.assertEquals(tdm.pageNumber, 2)
        self.assertEquals(tdm.totalItems, 107)
        self.assertEquals(tdm.totalPages, 8)
        tdm.nextPage()
        self.assertEquals(tdm.pageNumber, 3)
        self.assertEquals(tdm.totalItems, 107)
        self.assertEquals(tdm.totalPages, 8)
        tdm.nextPage()
        self.assertEquals(tdm.pageNumber, 4)
        self.assertEquals(tdm.totalItems, 107)
        self.assertEquals(tdm.totalPages, 8)
        tdm.nextPage()
        self.assertEquals(tdm.pageNumber, 5)
        self.assertEquals(tdm.totalItems, 107)
        self.assertEquals(tdm.totalPages, 8)
        tdm.nextPage()
        self.assertEquals(tdm.pageNumber, 6)
        self.assertEquals(tdm.totalItems, 107)
        self.assertEquals(tdm.totalPages, 8)
        tdm.nextPage()
        self.assertEquals(tdm.pageNumber, 7)
        self.assertEquals(tdm.totalItems, 107)
        self.assertEquals(tdm.totalPages, 8)
        tdm.nextPage()
        self.assertEquals(tdm.pageNumber, 8)
        self.assertEquals(tdm.totalItems, 107)
        self.assertEquals(tdm.totalPages, 8)

        tdm.lastPage()
        self.assertEquals(tdm.pageNumber, 8)
        self.assertEquals(tdm.totalPages, 8)
        self.assertEquals(tdm.totalItems, 107)
        tdm.firstPage()
        self.assertEquals(tdm.pageNumber, 1)
        self.assertEquals(tdm.totalPages, 8)
        self.assertEquals(tdm.totalItems, 107)




    def testSortItemUnderrun(self):
        tdm = tdb.TabularDataModel(self.store,
                X, [X.number,
                    X.textNumber,
                    X.phoneticDigits],
                itemsPerPage=110)

        assertFirstPage = lambda: self.assertNumbersAre(tdm, range(107))

        assertFirstPage()
        tdm.resort(tdm.currentSortColumn.attributeID)
        assertFirstPage()
        tdm.resort(tdm.currentSortColumn.attributeID, False)
        self.assertNumbersAre(tdm, list(reversed(range(107))))
        tdm.resort(tdm.currentSortColumn.attributeID)
        assertFirstPage()
        tdm.resort(tdm.currentSortColumn.attributeID, True)
        assertFirstPage()

    def testChangeSortColumnItemUnderrun(self):
        tdm = tdb.TabularDataModel(self.store,
                X, [X.number,
                    X.textNumber,
                    X.phoneticDigits],
                itemsPerPage=110)

        assertFirstPage = lambda: self.assertNumbersAre(tdm, range(107))

        assertFirstPage()
        tdm.resort(tdm.currentSortColumn.attributeID, True)
        assertFirstPage()
        # change sort direction, keep column
        tdm.resort(tdm.currentSortColumn.attributeID, False)
        self.assertNumbersAre(tdm, list(reversed(range(107))))
        # change sort column and direction
        tdm.resort('phoneticDigits', True)
        # switch back to previous sort column & direction
        tdm.resort('number', False)
        self.assertNumbersAre(tdm, list(reversed(range(107))))
        msg = 'itemsPerPage > totalItems, should only have one page'
        self.failIf(tdm.hasPrevPage(), msg)
        self.failIf(tdm.hasNextPage(), msg)

    def testOnePageHasNextPrev(self):
        bigtdm = tdb.TabularDataModel(self.store,
                X, [X.number,
                    X.textNumber,
                    X.phoneticDigits],
                    itemsPerPage=200)
        bigtdm.nextPage()
        bigtdm.nextPage()
        self.failIf(bigtdm.hasNextPage())
        self.failIf(bigtdm.hasPrevPage())

    def testMultiPageHasNextPrev(self):
        tdm = tdb.TabularDataModel(self.store,
                X, [X.number,
                    X.textNumber,
                    X.phoneticDigits],
                itemsPerPage=15)
        self.failIf(tdm.hasPrevPage())
        self.failUnless(tdm.hasNextPage())
        tdm.nextPage()
        self.failUnless(tdm.hasNextPage())
        self.failUnless(tdm.hasPrevPage())
        tdm.prevPage()
        self.failIf(tdm.hasPrevPage())
        self.failUnless(tdm.hasNextPage())
        tdm.lastPage()
        self.failIf(tdm.hasNextPage())
        self.failUnless(tdm.hasPrevPage())
        tdm.prevPage()
        self.failUnless(tdm.hasNextPage())
        self.failUnless(tdm.hasPrevPage())
        tdm.firstPage()
        for x in range(6):
            tdm.nextPage()
        self.failUnless(tdm.hasPrevPage())
        self.failUnless(tdm.hasNextPage())
        tdm.nextPage()
        self.failUnless(tdm.hasPrevPage())
        self.failIf(tdm.hasNextPage())

    def testCurrentPageDoesntChange(self):
        tdm = tdb.TabularDataModel(self.store,
                                   X, [X.number,
                                       X.textNumber,
                                       X.phoneticDigits],
                                   itemsPerPage=15)
        for y in range(3):
            tdm.nextPage()
            example = tdm.currentPage()
            for x in range(3):
                self.assertEquals(example, tdm.currentPage())


    def testDeleteItems(self):
        tdm = tdb.TabularDataModel(self.store,
                                   X, [X.number,
                                       X.textNumber,
                                       X.phoneticDigits],
                                   itemsPerPage=15)
        r = tdm.currentPage()
        r0 = r[0]['__item__']
        r0.deleteFromStore()
        rx = tdm.currentPage()
        r1 = rx[0]['__item__']
        self.failIf(r1._Item__deleting,
                    "First item in current page was deleted!")
        r1.deleteFromStore()
        tdm.nextPage()
        self.assertNumbersAre(tdm, range(16, 16+15))
        tdm.prevPage()
        self.assertNumbersAre(tdm, range(2, 2+15))

        # The distinction between this test and the nextPage test above is
        # important: we never "skip" items when paging.  The current results of
        # the TDM represent *what the user has seen*.  In the previous case, we
        # have never called currentPage and received a result with #16 in it
        # before calling nextPage.  In this case, assertNumbersAre(2,...) has
        # "seen" the page with #16 on it, so we are (correctly) taken to the
        # next item after that to begin the next page.
        tdm.nextPage()
        self.assertNumbersAre(tdm, range(17, 17+15))

    def testSorting(self):
        tdm = tdb.TabularDataModel(self.store,
                X, [X.number,
                    X.textNumber,
                    X.phoneticDigits],
                itemsPerPage=15)

        _assertNumbersAre = lambda seq: self.assertNumbersAre(tdm, seq)
        tdm.resort('number', True)
        _assertNumbersAre(range(15))
        tdm.resort('number', False)
        _assertNumbersAre(list(reversed(range(15))))
        _assertNumbersAre(list(reversed(range(15))))
        tdm.nextPage()
        _assertNumbersAre(list(reversed(range(15))))

    def testUnsortableColumns(self):
        tdm = tdb.TabularDataModel(self.store,
                X, [X.number,
                    UnsortableColumn(X.textNumber),
                    UnsortableColumn(X.phoneticDigits)],
                itemsPerPage=15)

        self.assertNumbersAre(tdm, range(15))
        self.assertRaises(scrolltable.Unsortable,
                          lambda: tdm.resort('textNumber'))
        self.assertRaises(scrolltable.Unsortable,
                          lambda: tdm.resort('phoneticDigits'))
        # check to see if the last valid state remains
        self.assertNumbersAre(tdm, range(15))




from nevow.athena import expose
from nevow.livetrial.testcase import TestCase

from axiom.item import Item
from axiom.store import Store
from axiom.attributes import integer
from axiom.dependency import installOn

from xmantissa.scrolltable import SequenceScrollingFragment
from xmantissa.webtheme import getLoader
from xmantissa.webapp import PrivateApplication


class ScrollElement(Item):
    """
    Dummy item used to populate scrolltables for the scrolltable tests.
    """
    column = integer()



class ScrollTableModelTestCase(TestCase):
    """
    Tests for the scrolltable's model class.
    """
    jsClass = u'Mantissa.Test.ScrollTableModelTestCase'



class ScrollTableWidgetTestCase(TestCase):
    """
    Tests for the scrolltable's view class.
    """
    jsClass = u'Mantissa.Test.ScrollTableViewTestCase'

    def __init__(self):
        TestCase.__init__(self)
        self.perTestData = {}


    def getScrollingWidget(self, key, rowCount=10):
        store = Store()
        installOn(PrivateApplication(store=store), store)
        elements = [ScrollElement(store=store, column=i) for i in range(rowCount)]
        columns = [ScrollElement.column]
        f = SequenceScrollingFragment(store, elements, columns)
        f.docFactory = getLoader(f.fragmentName)
        f.setFragmentParent(self)
        self.perTestData[key] = (store, elements, f)
        return f
    expose(getScrollingWidget)


    def changeRowCount(self, key, n):
        store, elements, fragment = self.perTestData[key]
        elements[:] = [ScrollElement(store=store, column=i) for i in range(n)]
    expose(changeRowCount)



class ScrollTableActionsTestCase(ScrollTableWidgetTestCase):
    """
    Tests for scrolltable actions
    """
    jsClass = u'Mantissa.Test.ScrollTableActionsTestCase'

    def getScrollingWidget(self, key, *a, **kw):
        f = ScrollTableWidgetTestCase.getScrollingWidget(self, key, *a, **kw)
        f.jsClass = u'Mantissa.Test.ScrollTableWithActions'

        # close over "key" because actions can't supply additional
        # arguments, and there isn't a use case outside of this test
        def action_delete(scrollElement):
            elements = self.perTestData[key][1]
            elements.remove(scrollElement)

        f.action_delete = action_delete
        return f
    expose(getScrollingWidget)



class ScrollTablePlaceholderRowsTestCase(ScrollTableWidgetTestCase):
    """
    Tests for the scrolltable's placeholder rows
    """
    jsClass = u'Mantissa.Test.ScrollTablePlaceholderRowsTestCase'




"""
Account configuration and management features, via the web.

This is a pitiful implementation of these concepts (hence the pitiful
module name).  It will be replaced by a real implementation when
clustering is ready for general use.
"""

import pytz

from zope.interface import implements

from twisted.python.components import registerAdapter
from twisted.internet import defer

from nevow import inevow, athena
from nevow.athena import expose

from epsilon import extime

from axiom import item, attributes, userbase

from xmantissa import ixmantissa, websession

class InvalidPassword(Exception):
    pass

class NonExistentAccount(Exception):
    pass

class NoSuchSession(Exception):
    pass


class AuthenticationApplication(item.Item):

    typeName = 'mantissa_web_authentication_application'
    schemaVersion = 1

    lastCredentialsChange = attributes.timestamp(allowNone=False)

    def __init__(self, **kw):
        if 'lastCredentialsChange' not in kw:
            kw['lastCredentialsChange'] = extime.Time()
        super(AuthenticationApplication, self).__init__(**kw)

    def _account(self):
        substore = self.store.parent.getItemByID(self.store.idInParent)
        for account in self.store.parent.query(userbase.LoginAccount,
                                               userbase.LoginAccount.avatars == substore):
            return account
        raise NonExistentAccount()

    def _username(self):
        for (localpart, domain) in userbase.getAccountNames(self.store):
            return (localpart + '@' + domain).encode('utf-8')

    def hasCurrentPassword(self):
        return defer.succeed(self._account().password is not None)

    def changePassword(self, oldPassword, newPassword):
        account = self._account()
        if account.password is not None and account.password != oldPassword:
            raise InvalidPassword()
        else:
            account.password = newPassword
            self.lastCredentialsChange = extime.Time()


    def persistentSessions(self):
        username = self._username()
        return self.store.parent.query(
            websession.PersistentSession,
            websession.PersistentSession.authenticatedAs == username)


    def cancelPersistentSession(self, uid):
        username = self._username()
        for sess in self.store.parent.query(websession.PersistentSession,
                                            attributes.AND(websession.PersistentSession.authenticatedAs == username,
                                                           websession.PersistentSession.sessionKey == uid)):
            sess.deleteFromStore()
            break
        else:
            raise NoSuchSession()


class AuthenticationFragment(athena.LiveFragment):
    implements(ixmantissa.INavigableFragment)

    fragmentName = 'authentication-configuration'
    live = 'athena'
    title = 'Change Password'

    jsClass = u'Mantissa.Authentication'

    def __init__(self, original):
        self.store = original.store
        athena.LiveFragment.__init__(self, original)

    def head(self):
        return None

    def render_currentPasswordField(self, ctx, data):
        d = self.original.hasCurrentPassword()

        def cb(result):
            if result:
                patName = 'current-password'
            else:
                patName = 'no-current-password'
            return inevow.IQ(self.docFactory).onePattern(patName)

        return d.addCallback(cb)


    def render_cancel(self, ctx, data):
        # XXX See previous XXX 19th
        handler = 'Mantissa.Authentication.get(this).cancel(%r); return false'
        return ctx.tag(onclick=handler % (data['session'].sessionKey,))


    def changePassword(self, currentPassword, newPassword):
        try:
            self.original.changePassword(unicode(currentPassword), unicode(newPassword))
        except NonExistentAccount:
            raise NonExistentAccount('You do not seem to exist.  Password unchanged.')
        except InvalidPassword:
            raise InvalidPassword('Incorrect password!  Nothing changed.')
        else:
            return u'Password Changed!'
    expose(changePassword)


    def cancel(self, uid):
        try:
            self.original.cancelPersistentSession(str(uid))
        except NoSuchSession:
            raise NoSuchSession('That session seems to have vanished.')
        else:
            return u'Session discontinued'
    expose(cancel)


    def data_persistentSessions(self, ctx, data):
        zone = pytz.timezone('US/Eastern')
        for session in self.original.persistentSessions():
            yield dict(lastUsed=session.lastUsed.asHumanly(zone) + ' ' + zone.zone,
                       session=session)


registerAdapter(AuthenticationFragment, AuthenticationApplication, ixmantissa.INavigableFragment)



# -*- test-case-name: xmantissa.test.historic.test_ampConfiguration1to2 -*-
# Copyright 2008 Divmod, Inc. See LICENSE file for details

"""
Generate a stub for the tests for the upgrade from schema version 1 to 2 of the
AMPConfiguration item.
"""

from axiom.dependency import installOn
from axiom.test.historic.stubloader import saveStub

from xmantissa.ampserver import AMPConfiguration


def createDatabase(store):
    """
    Make an L{AMPConfiguration} in the given store.
    """
    installOn(AMPConfiguration(store=store), store)


if __name__ == '__main__':
    saveStub(createDatabase, 16892)



"""
An interactive demonstration of various people-related functionality.

Run this test like this::
    $ twistd -n athena-widget --element=xmantissa.test.acceptance.people.{editperson,addperson}
    $ firefox http://localhost:8080/
"""

from axiom.store import Store
from axiom.dependency import installOn

from xmantissa.people import Organizer, EditPersonView, AddPersonFragment

store = Store()
organizer = Organizer(store=store)
installOn(organizer, store)
person = organizer.createPerson(u'alice')

def editperson():
    """
    Create a database with a Person in it and return the L{EditPersonView} for
    that person.
    """
    return EditPersonView(person)



def addperson():
    """
    Create a database with an L{Organizer} in it and return a
    L{AddPersonFragment} wrapped around the organizer.
    """
    return AddPersonFragment(organizer)




from xmantissa import prefs

def createDatabase(s):
    prefs.DefaultPreferenceCollection(store=s)

from axiom.test.historic.stubloader import saveStub

if __name__ == '__main__':
    saveStub(createDatabase)



from zope.interface import implements

from twisted.trial.unittest import TestCase

from axiom.item import Item
from axiom.store import Store
from axiom import attributes
from axiom.dependency import installOn

from xmantissa import prefs, ixmantissa, liveform

class WidgetShopPrefCollection(Item, prefs.PreferenceCollectionMixin):
    """
    Basic L{xmantissa.ixmantissa.IPreferenceCollection}, with a single
    preference, C{preferredWidget}
    """
    implements(ixmantissa.IPreferenceCollection)

    installedOn = attributes.reference()
    preferredWidget = attributes.text()
    powerupInterfaces = (ixmantissa.IPreferenceCollection,)

    def getPreferenceParameters(self):
        return (liveform.Parameter('preferredWidget',
                                    liveform.TEXT_INPUT,
                                    unicode,
                                    'Preferred Widget'),)

class PreferencesTestCase(TestCase):
    """
    Test case for basic preference functionality
    """

    def testAggregation(self):
        """
        Assert that L{xmantissa.prefs.PreferenceAggregator} gives us
        the right values for the preference attributes on L{WidgetShopPrefCollection}
        """
        store = Store()

        agg = prefs.PreferenceAggregator(store=store)
        installOn(agg, store)

        coll = WidgetShopPrefCollection(store=store)
        installOn(coll, store)

        coll.preferredWidget = u'Foo'

        self.assertEqual(agg.getPreferenceValue('preferredWidget'), 'Foo')

        coll.preferredWidget = u'Bar'

        self.assertEqual(agg.getPreferenceValue('preferredWidget'), 'Bar')


    def testGetPreferences(self):
        """
        Test that L{prefs.PreferenceCollectionMixin.getPreferences} works
        """
        class TrivialPreferenceCollection(prefs.PreferenceCollectionMixin):
            foo = 'bar'
            def getPreferenceParameters(self):
                return (liveform.Parameter('foo', liveform.TEXT_INPUT, str),)

        self.assertEqual(TrivialPreferenceCollection().getPreferences(),
                         {'foo': 'bar'})


    def testGetPreferencesNone(self):
        """
        Test that L{prefs.PreferenceCollectionMixin.getPreferences} does the
        right thing when the preference collection returns None from
        C{getPreferenceParameters}
        """
        class TrivialPreferenceCollection(prefs.PreferenceCollectionMixin):
            def getPreferenceParameters(self):
                return None

        self.assertEqual(TrivialPreferenceCollection().getPreferences(), {})




from twisted.trial import unittest

from axiom import store

from xmantissa import ixmantissa, endpoint

class MantissaQ2Q(unittest.TestCase):
    def testInstallation(self):
        d = self.mktemp()
        s = store.Store(unicode(d))
        q = endpoint.UniversalEndpointService(store=s)
        q.installOn(s)
        self.assertIdentical(ixmantissa.IQ2QService(s), q)




"""
Fulltext index a message a fixed number of times with PyLucene via the Mantissa
fulltext indexing API.
"""

from zope.interface import implements

from epsilon.scripts import benchmark

from axiom import store

from xmantissa import ixmantissa, fulltext


class Message(object):
    implements(ixmantissa.IFulltextIndexable)

    def uniqueIdentifier(self):
        return str(id(self))


    def textParts(self):
        return [
            u"Hello, how are you.  Please to be "
            u"seeing this message as an indexer test." * 100]


    def keywordParts(self):
            return {u'foo': "A Keyword"}


    def documentType(self):
        return u'message'


    def sortKey(self):
        return u''

def main():
    s = store.Store("lucene.axiom")
    indexer = fulltext.PyLuceneIndexer(store=s)

    benchmark.start()
    writer = indexer.openWriteIndex()
    for i in xrange(10000):
        writer.add(Message())
    writer.close()
    benchmark.stop()



if __name__ == '__main__':
    main()



from epsilon import setuphelper

from xmantissa import version

setuphelper.autosetup(
    name="Mantissa",
    version=version.short(),
    maintainer="Divmod, Inc.",
    maintainer_email="support@divmod.org",
    url="http://divmod.org/trac/wiki/DivmodMantissa",
    license="MIT",
    platforms=["any"],
    description="A multiprotocol application deployment platform",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: No Input/Output (Daemon)",
        "Framework :: Twisted",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: JavaScript",
        "Programming Language :: Python",
        "Topic :: Internet",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Terminals",
        ],
    )



"""
This module is here to satisfy old databases that contain MyAccount items
"""

from axiom.item import Item
from axiom.upgrade import registerUpgrader

class MyAccount(Item):
    typeName = 'mantissa_myaccount'
    schemaVersion = 2

def deleteMyAccount(old):
    # Just get rid of the old account object.  Don't even create a new one.
    old.deleteFromStore()
    return None

registerUpgrader(deleteMyAccount, 'mantissa_myaccount', 1, 2)



from axiom.test.historic import stubloader
from xmantissa.people import Organizer
from xmantissa.webapp import PrivateApplication

class OrganizerTest(stubloader.StubbedTest):
    def testUpgrade(self):
        o = self.store.findUnique(Organizer)
        self.failUnless(isinstance(o._webTranslator, PrivateApplication))




from twisted.trial.unittest import TestCase

from xmantissa.smtp import Address, parseAddress
from xmantissa.error import ArgumentError

class AddressParserTestCase(TestCase):
    """
    Test the RFC 2821 address parser.
    """
    def _addrTest(self, addr, exp):
        """
        Assert that the given address string parses to the given address
        object.

        @type addr: C{str}
        @type exp: L{Address}
        """
        a = parseAddress(addr)
        self.assertEquals(a, exp, "Misparsed %r to %r" % (addr, a))


    def test_longAddressRejected(self):
        """
        Test that an address longer than 256 bytes is rejected as illegal.
        """
        format = '<%s@example.com>'
        address = format % ('x' * (257 - len(format) + 2),)
        self.assertEqual(len(address), 257)
        self.assertRaises(ArgumentError, parseAddress, address)


    def test_nullAddress(self):
        """
        Test the parsing of the null address.
        """
        return self._addrTest('<>', Address(None, None, None))


    def test_emptyAddress(self):
        """
        Test the parsing of an address with empty local and domain parts.
        """
        return self._addrTest('<@>', Address(None, '', ''))


    def test_localOnly(self):
        """
        Test the parsing of an address with a non-empty local part and an
        empty domain part.
        """
        return self._addrTest('<localpart@>', Address(None, 'localpart', ''))


    def test_domainOnly(self):
        """
        Test the parsing of an address with an empty local part and a
        non-empty domain part.
        """
        return self._addrTest(
            '<@example.com>', Address(None, '', 'example.com'))


    def test_commonAddress(self):
        """
        Test the common case, an address with non-empty local and domain
        parts.
        """
        return self._addrTest(
            '<localpart@example.com>',
            Address(None, 'localpart', 'example.com'))


    def test_dottedLocalpart(self):
        """
        Test parsing of an address with a dotted local part.
        """
        return self._addrTest(
            '<local.part@>', Address(None, 'local.part', ''))


    def test_ipv4Literal(self):
        """
        Test parsing of an IPv4 domain part literal.
        """
        return self._addrTest('<@9.8.7.6>', Address(None, '', '9.8.7.6'))


    def test_ipv4LiteralWithLocalpart(self):
        """
        Test parsing of an IPv4 domain part literal with a non-empty local
        part.
        """
        return self._addrTest(
            '<localpart@1.2.3.4>', Address(None, 'localpart', '1.2.3.4'))


    def test_singlePartSourceRoute(self):
        """
        Test parsing of an address with a one element source route.
        """
        return self._addrTest(
            '<@foo.bar:localpart@example.com>',
            Address(['foo.bar'], 'localpart', 'example.com'))


    def test_multiplePartSourceRoute(self):
        """
        Test parsing of an address with a two element source route.
        """
        return self._addrTest(
            '<@foo.bar,@bar.baz:localpart@example.com>',
            Address(['foo.bar', 'bar.baz'], 'localpart', 'example.com'))


    def test_ipv6Literal(self):
        """
        Test parsing of an IPv6 domain part literal.
        """
        return self._addrTest('<@IPv6:::1>', Address(None, '', '::1'))
    test_ipv6Literal.skip = "Add IPv6 support to Twisted"


    def test_ipv6LiteralWithLocalpart(self):
        """
        Test parsing of an IPv6 domain part literal with a non-empty local
        part.
        """
        return self._addrTest(
            '<localpart@IPv6:::1>', Address(None, 'localpart', '::1'))
    test_ipv6LiteralWithLocalpart.skip = "Add IPv6 support to Twisted"



"""
Database-creation script for testing the version 1 to version 2 upgrader of
L{xmantissa.publicweb.FrontPage}.
"""
from axiom.test.historic.stubloader import saveStub
from xmantissa.publicweb import FrontPage

def createDatabase(s):
    """
    Create a L{FrontPage} item.
    """
    FrontPage(store=s, publicViews=17, privateViews=42)

if __name__ == '__main__':
    saveStub(createDatabase, 16183)





from axiom.test.historic.stubloader import StubbedTest

from xmantissa.search import SearchResult

class SearchResultUpgradeTestCase(StubbedTest):
    def test_removal(self):
        """
        SearchResults are no longer necessary, so the upgrade to version two
        should delete them completely.
        """
        self.assertEquals(
            list(self.store.query(SearchResult)),
            [])



# -*- test-case-name: xmantissa.test.test_webcmd -*-

import os
import sys

from twisted.python import reflect
from twisted.python.usage import UsageError

from axiom import item, attributes
from axiom.dependency import installOn, onlyInstallPowerups
from axiom.scripts import axiomatic

from xmantissa.web import SiteConfiguration
from xmantissa.website import StaticSite, APIKey
from xmantissa import ixmantissa, webadmin
from xmantissa.plugins.baseoff import baseOffering


class WebConfiguration(axiomatic.AxiomaticCommand):
    name = 'web'
    description = 'Web.  Yay.'

    optParameters = [
        ('http-log', 'h', None,
         'Filename (relative to files directory of the store) to which to log '
         'HTTP requests (empty string to disable)'),
        ('hostname', 'H', None,
         'Canonical hostname for this server (used in URL generation).'),
        ('urchin-key', '', None,
         'Google Analytics API key for this site')]

    def __init__(self, *a, **k):
        super(WebConfiguration, self).__init__(*a, **k)
        self.staticPaths = []


    didSomething = 0

    def postOptions(self):
        siteStore = self.parent.getStore()

        # Make sure the base mantissa offering is installed.
        offeringTech = ixmantissa.IOfferingTechnician(siteStore)
        offerings = offeringTech.getInstalledOfferingNames()
        if baseOffering.name not in offerings:
            raise UsageError(
                "This command can only be used on Mantissa databases.")

        # It is, we can make some simplifying assumptions.  Specifically,
        # there is exactly one SiteConfiguration installed.
        site = siteStore.findUnique(SiteConfiguration)

        if self['http-log'] is not None:
            if self['http-log']:
                site.httpLog = siteStore.filesdir.preauthChild(
                    self['http-log'])
            else:
                site.httpLog = None

        if self['hostname'] is not None:
            if self['hostname']:
                site.hostname = self.decodeCommandLine(self['hostname'])
            else:
                raise UsageError("Hostname may not be empty.")

        if self['urchin-key'] is not None:
            # Install the API key for Google Analytics, to enable tracking for
            # this site.
            APIKey.setKeyForAPI(
                siteStore, APIKey.URCHIN, self['urchin-key'].decode('ascii'))


        # Set up whatever static content was requested.
        for webPath, filePath in self.staticPaths:
            staticSite = siteStore.findFirst(
                StaticSite, StaticSite.prefixURL == webPath)
            if staticSite is not None:
                staticSite.staticContentPath = filePath
            else:
                staticSite = StaticSite(
                    store=siteStore,
                    staticContentPath=filePath,
                    prefixURL=webPath,
                    sessionless=True)
                onlyInstallPowerups(staticSite, siteStore)


    def opt_static(self, pathMapping):
        webPath, filePath = self.decodeCommandLine(pathMapping).split(os.pathsep, 1)
        if webPath.startswith('/'):
            webPath = webPath[1:]
        self.staticPaths.append((webPath, os.path.abspath(filePath)))


    def opt_list(self):
        self.didSomething = 1
        s = self.parent.getStore()
        for ws in s.query(SiteConfiguration):
            print 'The hostname is', ws.hostname
            if ws.httpLog is not None:
                print 'Logging HTTP requests to', ws.httpLog
            break
        else:
            print 'No configured webservers.'


        def powerupsWithPriorityFor(interface):
            for cable in s.query(
                item._PowerupConnector,
                attributes.AND(item._PowerupConnector.interface == unicode(reflect.qual(interface)),
                               item._PowerupConnector.item == s),
                sort=item._PowerupConnector.priority.descending):
                yield cable.powerup, cable.priority

        print 'Sessionless plugins:'
        for srp, prio in powerupsWithPriorityFor(ixmantissa.ISessionlessSiteRootPlugin):
            print '  %s (prio. %d)' % (srp, prio)
        print 'Sessioned plugins:'
        for srp, prio in powerupsWithPriorityFor(ixmantissa.ISiteRootPlugin):
            print '  %s (prio. %d)' % (srp, prio)
        sys.exit(0)

    opt_static.__doc__ = """
    Add an element to the mapping of web URLs to locations of static
    content on the filesystem (webpath%sfilepath)
    """ % (os.pathsep,)



class WebAdministration(axiomatic.AxiomaticCommand):
    name = 'web-admin'
    description = 'Administrative controls for the web'

    optFlags = [
        ('admin', 'a', 'Enable administrative controls'),
        ('developer', 'd', 'Enable developer controls'),

        ('disable', 'D', 'Remove the indicated options, instead of enabling them.'),
        ]

    def postOptions(self):
        s = self.parent.getStore()

        didSomething = False

        if self['admin']:
            didSomething = True
            if self['disable']:
                for app in s.query(webadmin.AdminStatsApplication):
                    app.deleteFromStore()
                    break
                else:
                    raise UsageError('Administrator controls already disabled.')
            else:
                installOn(webadmin.AdminStatsApplication(store=s), s)

        if self['developer']:
            didSomething = True
            if self['disable']:
                for app in s.query(webadmin.DeveloperApplication):
                    app.deleteFromStore()
                    break
                else:
                    raise UsageError('Developer controls already disabled.')
            else:
                installOn(webadmin.DeveloperApplication(store=s), s)

        if not didSomething:
            raise UsageError("Specify something or I won't do anything.")




from xmantissa import signup

freeTicket = signup.SignupMechanism(
    name = 'Free Ticket',
    description = '''
    Create a page which will allow anyone with a verified email
    address to sign up for the system.  When the user enters their
    email address, a confirmation email is sent to it containing a
    link which will allow signup to proceed.  When the link is
    followed, an account will be created and endowed by the
    benefactors associated with this instance.
    ''',
    itemClass = signup.FreeTicketSignup,
    configuration = signup.freeTicketSignupConfiguration)

userInfo = signup.SignupMechanism(
    name = 'Required User Information',
    description = '''
    Create a signup mechanism with several self-validating fields.

    This will also require the user to select a local username before the
    account is created, and it will create the account immediately rather than
    waiting for the ticket to be claimed.
    ''',
    itemClass = signup.UserInfoSignup,
    configuration = signup.freeTicketSignupConfiguration)




import sys, os, struct

from zope.interface import directlyProvides

from twisted.python import util
from twisted.cred import portal
from twisted.plugin import IPlugin

from axiom import errors as eaxiom
from axiom.scripts import axiomatic
from axiom.attributes import AND
from axiom.dependency import installOn
from axiom.iaxiom import IVersion

from xmantissa.ixmantissa import IOfferingTechnician
from xmantissa import webadmin, publicweb, stats
from xmantissa.web import SiteConfiguration
from xmantissa.terminal import SecureShellConfiguration
from xmantissa.port import TCPPort, SSLPort
from xmantissa.plugins.baseoff import baseOffering

# PortConfiguration isn't used here, but it's a plugin, so it gets discovered
# here.
from xmantissa.port import PortConfiguration
#version also gets registered as a plugin here.
from xmantissa import version

from epsilon.asplode import splode
from epsilon.scripts import certcreate

directlyProvides(version, IPlugin, IVersion)



def gtpswd(prompt, confirmPassword):
    """
    Temporary wrapper for Twisted's getPassword until a version that supports
    customizing the 'confirm' prompt is released.
    """
    try:
        return util.getPassword(prompt=prompt,
                                confirmPrompt=confirmPassword,
                                confirm=True)
    except TypeError:
        return util.getPassword(prompt=prompt,
                                confirm=True)



def genSerial():
    """
    Generate a (hopefully) unique integer usable as an SSL certificate serial.
    """
    return abs(struct.unpack('!l', os.urandom(4))[0])



class Mantissa(axiomatic.AxiomaticCommand):
    """
    Create all the moving parts necessary to begin interactively developing a
    Mantissa application component of your own.
    """

    # Throughout here we use findOrCreate rather than raw creation so that
    # duplicate installations of these components do not create garbage
    # objects.

    # Yea?  Where's the unit tests for that?  And who re-runs "axiomatic
    # mantissa" on the same Store multiple times anyway?  Better that it should
    # give an error than transparently... do... something... maybe...  I'm
    # actively not preserving the idempotency behavior in my changes to this
    # code. -exarkun

    name = 'mantissa'
    description = 'Blank Mantissa service'

    longdesc = __doc__

    optParameters = [
        ('admin-user', 'a', 'admin@localhost',
         'Account name for the administrative user.'),
        ('admin-password', 'p', None,
         'Password for the administrative user '
         '(if omitted, will be prompted for).'),
        ('public-url', None, '',
         'URL at which to publish the public front page.')]

    def postOptions(self):
        siteStore = self.parent.getStore()
        if self['admin-password'] is None:
            pws = u'Divmod\u2122 Mantissa\u2122 password for %r: ' % (self['admin-user'],)
            self['admin-password'] = gtpswd((u'Enter ' + pws).encode(sys.stdout.encoding, 'ignore'),
                                            (u'Confirm ' + pws).encode(sys.stdout.encoding, 'ignore'))


        publicURL = self.decodeCommandLine(self['public-url'])
        adminUser = self.decodeCommandLine(self['admin-user'])
        adminPassword = self['admin-password']

        adminLocal, adminDomain = adminUser.split(u'@')

        siteStore.transact(self.installSite, siteStore, adminDomain, publicURL)
        siteStore.transact(
            self.installAdmin, siteStore, adminLocal, adminDomain, adminPassword)


    def installSite(self, siteStore, domain, publicURL, generateCert=True):
        """
        Create the necessary items to run an HTTP server and an SSH server.
        """
        certPath = siteStore.filesdir.child("server.pem")
        if generateCert and not certPath.exists():
            certcreate.main([
                    '--filename', certPath.path, '--quiet',
                    '--serial-number', str(genSerial()),
                    '--hostname', domain])

        # Install the base Mantissa offering.
        IOfferingTechnician(siteStore).installOffering(baseOffering)

        # Make the HTTP server baseOffering includes listen somewhere.
        site = siteStore.findUnique(SiteConfiguration)
        site.hostname = domain
        installOn(
            TCPPort(store=siteStore, factory=site, portNumber=8080),
            siteStore)
        installOn(
            SSLPort(store=siteStore, factory=site, portNumber=8443,
                    certificatePath=certPath),
            siteStore)

        # Make the SSH server baseOffering includes listen somewhere.
        shell = siteStore.findUnique(SecureShellConfiguration)
        installOn(
            TCPPort(store=siteStore, factory=shell, portNumber=8022),
            siteStore)

        # Install a front page on the top level store so that the
        # developer will have something to look at when they start up
        # the server.
        fp = siteStore.findOrCreate(publicweb.FrontPage, prefixURL=u'')
        installOn(fp, siteStore)


    def installAdmin(self, s, username, domain, password):
        # Add an account for our administrator, so they can log in through the
        # web.
        r = portal.IRealm(s)
        try:
            acc = r.addAccount(username, domain, password, internal=True, verified=True)
        except eaxiom.DuplicateUser:
            acc = r.accountByAddress(username, domain)

        accStore = acc.avatars.open()
        accStore.transact(webadmin.endowAdminPowerups, accStore)



class Generate(axiomatic.AxiomaticCommand):
    name = "project"

    # This will show up next to the name in --help output
    description = "Generate most basic skeleton of a Mantissa app"

    optParameters = [
        ('name', 'n', None, 'The name of the app to deploy'),
        ]

    def postOptions(self):
        if self['name'] is None:
            proj = ''
            while( proj == ''):
                try:
                    proj = raw_input("Please provide the name of the app to deploy: " )
                except KeyboardInterrupt:
                    raise SystemExit()
        else:
            proj = self.decodeCommandLine(self['name'])
        proj = proj.lower()
        capproj = proj.capitalize()
        print "Creating", capproj, "in", capproj

        fObj = file(util.sibpath(__file__, 'template.txt'))

        splode(fObj.readlines(), proj, capproj)



class RemoteStatsAdd(axiomatic.AxiomaticSubCommand):

    optParameters = [
        ("host", "h", None, "The host accepting statistical data."),
        ("port", "p", None, "The port to connect to."),
        ]

    def postOptions(self):
        s = self.parent.parent.getStore()
        s.transact(self.installCollector, s, self['host'], int(self['port']))


    def installCollector(self, s, host, port):
        ss = portal.IRealm(s).accountByAddress(u'mantissa',
                                               None).avatars.open()
        stats.RemoteStatsObserver(store=ss, hostname=host, port=port)



class RemoteStatsList(axiomatic.AxiomaticSubCommand):
    def postOptions(self):
        s = self.parent.parent.getStore()
        ss = portal.IRealm(s).accountByAddress(u'mantissa',
                                               None).avatars.open()
        for i, obs in enumerate(ss.query(stats.RemoteStatsObserver)):
            print "%s) %s:%s" % (i, obs.hostname, obs.port)



class RemoteStatsRemove(axiomatic.AxiomaticSubCommand):
    optParameters = [
        ("host", "h", None, "The hostname of the observer to remove."),
        ("port", "p", None, "The port of the observer to remove."),
        ]
    def postOptions(self):
        s = self.parent.parent.getStore()
        ss = portal.IRealm(s).accountByAddress(u'mantissa',
                                               None).avatars.open()
        for obs in ss.query(stats.RemoteStatsObserver,
                            AND(stats.RemoteStatsObserver.hostname==self['host'], stats.RemoteStatsObserver.port==int(self['port']))):
            obs.deleteFromStore()



class RemoteStats(axiomatic.AxiomaticCommand):
    name = "stats"
    description = "Control remote statistics collection"

    subCommands = [("add", None, RemoteStatsAdd, "Submit Mantissa statistical data to another server"),
                   ("list", None, RemoteStatsList, "List remote targets for stats delivery"),
                   ("remove", None, RemoteStatsRemove, "Remove a remote stats target")]


__all__ = [
    PortConfiguration.__name__, Mantissa.__name__, Generate.__name__,
    RemoteStats.__name__, RemoteStatsAdd.__name__, RemoteStatsList.__name__,
    RemoteStatsRemove.__name__]



from axiom.test.historic.stubloader import saveStub
from xmantissa.webadmin import DeveloperApplication

def createDatabase(s):
    DeveloperApplication(store=s)

if __name__ == '__main__':
    saveStub(createDatabase, 10876)




"""
Tests for the upgrade from schema version 1 to 2 of RealName, which deletes the
item.
"""

from axiom.test.historic.stubloader import StubbedTest

from xmantissa.people import Person, RealName


class RealNameUpgradeTests(StubbedTest):
    def test_deleted(self):
        """
        The L{RealName} should no longer exist in the database.
        """
        self.assertEqual(self.store.query(RealName).count(), 0)



"""
Upgrader tests for L{xmantissa.port} items.
"""

from xmantissa.port import TCPPort, SSLPort
from xmantissa.web import SiteConfiguration

from axiom.test.historic.stubloader import StubbedTest

from xmantissa.test.historic.stub_port1to2 import TCP_PORT, SSL_PORT

class PortInterfaceUpgradeTest(StubbedTest):
    """
    Schema upgrade tests for L{xmantissa.port} items.

    This upgrade adds an "interface" attribute.
    """
    def test_TCPPort(self):
        """
        Test the TCPPort 1->2 schema upgrade.
        """
        port = self.store.findUnique(TCPPort)
        self.assertEqual(port.portNumber, TCP_PORT)
        self.assertTrue(isinstance(port.factory, SiteConfiguration))
        self.assertEqual(port.interface, u'')

    def test_SSLPort(self):
        """
        Test the SSLPort 1->2 schema upgrade.
        """
        port = self.store.findUnique(SSLPort)
        self.assertEqual(port.portNumber, SSL_PORT)
        self.assertEqual(port.certificatePath,
                self.store.newFilePath('certificate'))
        self.assertTrue(isinstance(port.factory, SiteConfiguration))
        self.assertEqual(port.interface, u'')



# -*- test-case-name: xmantissa.test.test_stats -*-
# Copyright 2008 Divmod, Inc.  See LICENSE file for details

"""
Statistics collection and recording facility.

This is a system for sending locally recorded stats data to remote observers
for further processing or storage.

Example::
  twisted.python.log.msg(interface=axiom.iaxiom.IStatEvent, foo=1, bar=2, ...)
"""

from zope.interface import implements

from twisted.application import service
from twisted.protocols import policies
from twisted.python import log
from twisted.protocols.amp import AMP, Command, AmpList, Unicode

from axiom import iaxiom, item, attributes, upgrade

from xmantissa.ixmantissa import IBoxReceiverFactory


class StatUpdate(Command):
    """
    L{StatUpdate} is sent by the server to clients connected to it each time a
    stat log event is observed.  The command contains the information from the
    event.
    """
    requiresAnswer = False

    arguments = [
        ('data', AmpList([("key", Unicode()), ("value", Unicode())])),
        ]



class RemoteStatsCollectorFactory(item.Item):
    """
    A collector of stats for things that are remote (factory).
    """
    powerupInterfaces = (IBoxReceiverFactory,)
    implements(*powerupInterfaces)

    protocol = u"http://divmod.org/ns/mantissa-stats"

    _unused = attributes.integer(
        doc="""
        meaningless attribute, only here to satisfy Axiom requirement for at
        least one attribute.
        """)

    def getBoxReceiver(self):
        """
        Create a new AMP protocol for delivering stats information to a remote
        process.
        """
        return RemoteStatsCollector()



class RemoteStatsCollector(AMP):
    """
    An AMP protocol for sending stats to remote observers.
    """
    def startReceivingBoxes(self, sender):
        """
        Start observing log events for stat events to send.
        """
        AMP.startReceivingBoxes(self, sender)
        log.addObserver(self._emit)


    def stopReceivingBoxes(self, reason):
        """
        Stop observing log events.
        """
        AMP.stopReceivingBoxes(self, reason)
        log.removeObserver(self._emit)


    def _emit(self, event):
        """
        If the given event is a stat event, send a I{StatUpdate} command.
        """
        if (event.get('interface') is not iaxiom.IStatEvent and
            'athena_send_messages' not in event and
            'athena_received_messages' not in event):
            return

        out = []
        for k, v in event.iteritems():
            if k in ('system', 'message', 'interface', 'isError'):
                continue
            if not isinstance(v, unicode):
                v = str(v).decode('ascii')
            out.append(dict(key=k.decode('ascii'), value=v))
        self.callRemote(StatUpdate, data=out)



class BandwidthMeasuringProtocol(policies.ProtocolWrapper):
    """
    Wraps a Protocol and sends bandwidth stats to a BandwidthMeasuringFactory.
    """

    def write(self, data):
        self.factory.registerWritten(len(data))
        policies.ProtocolWrapper.write(self, data)


    def writeSequence(self, seq):
        self.factory.registerWritten(sum(map(len, seq)))
        policies.ProtocolWrapper.writeSequence(self, seq)


    def dataReceived(self, data):
        self.factory.registerRead(len(data))
        policies.ProtocolWrapper.dataReceived(self, data)



class BandwidthMeasuringFactory(policies.WrappingFactory):
    """
    Collects stats on the number of bytes written and read by protocol
    instances from the wrapped factory.
    """

    protocol = BandwidthMeasuringProtocol

    def __init__(self, wrappedFactory, protocolName):
        policies.WrappingFactory.__init__(self, wrappedFactory)
        self.name = protocolName


    def registerWritten(self, length):
        log.msg(interface=iaxiom.IStatEvent, **{"stat_bandwidth_" + self.name + "_up": length})


    def registerRead(self, length):
        log.msg(interface=iaxiom.IStatEvent, **{"stat_bandwidth_" + self.name + "_down": length})


__all__ = [
    'StatUpdate', 'RemoteStatsCollectorFactory', 'RemoteStatsCollector',
    ]

# Older stuff, mostly deprecated or not useful, you probably don't want to look
# below here. -exarkun

statDescriptions = {
    "page_renders": "Nevow page renders per minute",
    "messages_grabbed": "POP3 messages grabbed per minute",
    "messagesSent": "SMTP messages sent per minute",
    "messagesReceived": "SMTP messages received per minute",
    "mimePartsCreated": "MIME parts created per minute",
    "cache_hits": "Axiom cache hits per minute",
    "cursor_execute_time": "Seconds spent in cursor.execute per minute",
    "cursor_blocked_time": ("Seconds spent waiting for the database lock per "
                            "minute"),
    "commits": "Axiom commits per minute",
    "cache_misses": "Axiom cache misses per minute",
    "autocommits": "Axiom autocommits per minute",
    "athena_messages_sent": "Athena messages sent per minute",
    "athena_messages_received": "Athena messages received per minute",

    "actionDuration": "Seconds/Minute spent executing Imaginary Commands",
    "actionExecuted": "Imaginary Commands/Minute executed",

    "bandwidth_http_up": "HTTP KB/sec received",
    "bandwidth_http_down": "HTTP KB/sec sent",
    "bandwidth_https_up": "HTTPS KB/sec sent",
    "bandwidth_https_down": "HTTPS KB/sec received",
    "bandwidth_pop3_up": "POP3 server KB/sec sent",
    "bandwidth_pop3_down":"POP3 server KB/sec received",
    "bandwidth_pop3s_up":"POP3S server KB/sec sent",
    "bandwidth_pop3s_down": "POP3S server KB/sec received",
    "bandwidth_smtp_up": "SMTP server KB/sec sent",
    "bandwidth_smtp_down": "SMTP server KB/sec received",
    "bandwidth_smtps_up": "SMTPS server KB/sec sent",
    "bandwidth_smtps_down": "SMTPS server KB/sec received",
    "bandwidth_pop3-grabber_up": "POP3 grabber KB/sec sent",
    "bandwidth_pop3-grabber_down": "POP3 grabber KB/sec received",
    "bandwidth_sip_up": "SIP KB/sec sent",
    "bandwidth_sip_down": "SIP KB/sec received",
    "bandwidth_telnet_up": "Telnet KB/sec sent",
    "bandwidth_telnet_down": "Telnet KB/sec received",
    "bandwidth_ssh_up": "SSH KB/sec sent",
    "bandwidth_ssh_down": "SSH KB/sec received",

    "Imaginary logins": "Imaginary Logins/Minute",
    "Web logins": "Web Logins/Minute",
    "SMTP logins": "SMTP Logins/Minute",
    "POP3 logins": "POP3 Logins/Minute",
    }


class StatBucket(item.Item):
    """
    Obsolete.  Only present for schema compatibility.  Do not use.
    """
    schemaVersion = 2
    type = attributes.text(doc="A stat name, such as 'messagesReceived'")
    value = attributes.ieee754_double(default=0.0, doc='Total number of events for this time period')
    interval = attributes.text(doc='A time period, e.g. "quarter-hour" or "minute" or "day"')
    index = attributes.integer(doc='The position in the round-robin list for non-daily stats')
    time = attributes.timestamp(doc='When this bucket was last updated')
    attributes.compoundIndex(interval, type, index)
    attributes.compoundIndex(index, interval)


class QueryStatBucket(item.Item):
    """
    Obsolete.  Only present for schema compatibility.  Do not use.
    """
    type = attributes.text("the SQL query string")
    value = attributes.ieee754_double(default=0.0, doc='Total number of events for this time period')
    interval = attributes.text(doc='A time period, e.g. "quarter-hour" or "minute" or "day"')
    index = attributes.integer(doc='The position in the round-robin list for non-daily stats')
    time = attributes.timestamp(doc='When this bucket was last updated')
    attributes.compoundIndex(interval, type, index)


class StatSampler(item.Item):
    """
    Obsolete.  Only present for schema compatibility.  Do not use.
    """
    service = attributes.reference()

    def run(self):
        """
        Obsolete.  Only present to prevent errors in existing systems.  Do not
        use.
        """



class StatsService(item.Item, service.Service):
    """
    Obsolete.  Only present for schema compatibility.  Do not use.
    """
    installedOn = attributes.reference()
    parent = attributes.inmemory()
    running = attributes.inmemory()
    name = attributes.inmemory()
    statoscope = attributes.inmemory()
    queryStatoscope = attributes.inmemory()
    statTypes = attributes.inmemory()
    currentMinuteBucket = attributes.integer(default=0)
    currentQuarterHourBucket = attributes.integer(default=0)
    observers = attributes.inmemory()
    loginInterfaces = attributes.inmemory()
    userStats = attributes.inmemory()

    powerupInterfaces = (service.IService,)



class RemoteStatsObserver(item.Item):
    """
    Obsolete.  Only present for schema compatibility.  Do not use.
    """

    hostname = attributes.bytes(doc="A host to send stat updates to")
    port = attributes.integer(doc="The port to send stat updates to")
    protocol = attributes.inmemory(doc="The juice protocol instance to send stat updates over")



def upgradeStatBucket1to2(bucket):
    bucket.deleteFromStore()
upgrade.registerUpgrader(upgradeStatBucket1to2, 'xmantissa_stats_statbucket', 1, 2)



# Copyright (c) 2008 Divmod, Inc.
# See LICENSE for details.

"""
Tests for L{xmantissa.port}.
"""

import sys
import os
from StringIO import StringIO

from zope.interface import implements
from zope.interface.verify import verifyObject

from twisted.trial.unittest import TestCase
from twisted.python.filepath import FilePath
from twisted.application.service import IService, IServiceCollection
from twisted.internet.protocol import ServerFactory
from twisted.internet.defer import Deferred
from twisted.internet.ssl import CertificateOptions

from axiom.iaxiom import IAxiomaticCommand
from axiom.store import Store
from axiom.item import Item
from axiom.attributes import inmemory, integer
from axiom.dependency import installOn
from axiom.scripts.axiomatic import Options as AxiomaticOptions
from axiom.test.util import CommandStub

from xmantissa.ixmantissa import IProtocolFactoryFactory
from xmantissa.port import TCPPort, SSLPort
from xmantissa.port import PortConfiguration


CERTIFICATE_DATA = """
-----BEGIN CERTIFICATE-----
MIICmTCCAgICAQEwDQYJKoZIhvcNAQEEBQAwgZQxCzAJBgNVBAYTAlVTMRQwEgYD
VQQDEwtleGFtcGxlLmNvbTERMA8GA1UEBxMITmV3IFlvcmsxEzARBgNVBAoTCkRp
dm1vZCBMTEMxETAPBgNVBAgTCE5ldyBZb3JrMSIwIAYJKoZIhvcNAQkBFhNzdXBw
b3J0QGV4YW1wbGUuY29tMRAwDgYDVQQLEwdUZXN0aW5nMB4XDTA2MTIzMDE5MDEx
NloXDTA3MTIzMDE5MDExNlowgZQxCzAJBgNVBAYTAlVTMRQwEgYDVQQDEwtleGFt
cGxlLmNvbTERMA8GA1UEBxMITmV3IFlvcmsxEzARBgNVBAoTCkRpdm1vZCBMTEMx
ETAPBgNVBAgTCE5ldyBZb3JrMSIwIAYJKoZIhvcNAQkBFhNzdXBwb3J0QGV4YW1w
bGUuY29tMRAwDgYDVQQLEwdUZXN0aW5nMIGfMA0GCSqGSIb3DQEBAQUAA4GNADCB
iQKBgQCrmNNyXLHAETcDH8Uxhmbo8IhFFMx1C4i7oTHTKsmD84E3YFj/RdByrWrG
TL4XskALpfmw1+LxQmMO8n4sIsN3QmjkAWhFhMEquKv6NNN+sRo6vF+ytEasuYn/
7gY/iT7LYqUmKWckBsPYzT9elyOXi6miI0tFdeyfXRSxOslKewIDAQABMA0GCSqG
SIb3DQEBBAUAA4GBABotNizqPoGWIG5BMsl8lxseqiw/8AwvoiQNpYTrC8W+Umsg
oZEaMuVkf/NDJEa3TXdYcAzkFwGN9Cn/WCgHEkLxIZ66aHV0bfcE7YJjHRDrrLiY
chPndOGGrD3iTuWaGnauUcsjJ+RsxqHMBu6NRQYgkoYNsOr0UA1ek7O1vjMy
-----END CERTIFICATE-----
"""

PRIVATEKEY_DATA = """
-----BEGIN RSA PRIVATE KEY-----
MIICXAIBAAKBgQCrmNNyXLHAETcDH8Uxhmbo8IhFFMx1C4i7oTHTKsmD84E3YFj/
RdByrWrGTL4XskALpfmw1+LxQmMO8n4sIsN3QmjkAWhFhMEquKv6NNN+sRo6vF+y
tEasuYn/7gY/iT7LYqUmKWckBsPYzT9elyOXi6miI0tFdeyfXRSxOslKewIDAQAB
AoGAHd9YCBOs+gPFMO0J9iowpiKhhm0tfr7ISemw89MCC8+LUimatK3hsOURrn3T
peppDd4SDsA2iMuG1SZP4r0Wi9ZncZ+uj6KfVHg6rJZRDW2cPsGNyBw2HO8pFxnh
NsfxioutzCqJ9A0KwqSNQsBpOAlRWzP13+/W5wYAGK+yrLECQQDYgOhVR+1KOhty
CI0NVITNFL5IOZ254Eu46qbEGwPNJvkzdp+Wx5gsfCip9aiZgw3LMEeGXu9P1C4N
AqDM4uozAkEAyua0F0nCRLzjLAAw4odC+vA6jnq6K4M7QT6cQVwmrxgOj6jGEOGu
eaoWbXi2bKcxOGBNDZW0PVKmpq4hZblmmQJBALwFP0AIxg+HZRxkRrMD6oz77cBl
oQ+ytbAywH9ggq2gohzKcRAN6J8BeIMZn8EpqkoCdKtCOQyX1SJhXOpySjcCQDds
mZka7tQz/KISU0gtxqAhav1sjNpB+Lez0J8R+wctPR0E70XBQBW/3mx84uf/K7TI
qYOidx+hKiCxxDGzWVECQHNVutQ1ABjmv6EDJTo28QQsm5hNbfS+tVY3bSihNjLM
Y+O7ib90LsqfQ8r0GUphQVi4EA4QMJqaF7ZxKms79qA=
-----END RSA PRIVATE KEY-----
"""



class DummyPort(object):
    """
    Stub class used to track what reactor listen calls have been made and what
    created ports have been stopped.
    """
    stopping = None

    def __init__(self, portNumber, factory, contextFactory=None, interface=''):
        self.portNumber = portNumber
        self.factory = factory
        self.contextFactory = contextFactory
        self.interface = interface


    def stopListening(self):
        assert self.stopping is None
        self.stopping = Deferred()
        return self.stopping



class DummyFactory(Item):
    """
    Helper class used as a stand-in for a real protocol factory by the unit
    tests.
    """
    implements(IProtocolFactoryFactory)

    dummyAttribute = integer(doc="""
    Meaningless attribute which serves only to make this a valid Item subclass.
    """)

    realFactory = inmemory(doc="""
    A reference to the protocol factory which L{getFactory} will return.
    """)

    def getFactory(self):
        return self.realFactory



class PortTestsMixin:
    """
    Test method-defining mixin class for port types with C{portNumber} and
    C{factory} attributes.

    Included are tests for various persistence-related behaviors as well as the
    L{IService} implementation which all ports should have.

    @ivar portType: The L{Item} subclass which will be tested.

    @ivar lowPortNumber: A port number which requires privileges to bind on
    POSIX.  Used to test L{privilegedStartService}.

    @ivar highPortNumber: A port number which does not require privileges to
    bind on POSIX.  Used to test the interaction between
    L{privilegedStartService} and L{startService}.

    @ivar dbdir: The path at which to create the test L{Store}.  This must be
    bound before L{setUp} is run, since that is the only method which examines
    its value.

    @ivar ports: A list of ports which have been bound using L{listen}.
    created in L{setUp}.
    """
    portType = None

    lowPortNumber = 123
    highPortNumber = 1234
    someInterface = u'127.0.0.1'

    def port(self, **kw):
        """
        Create and return a new port instance with the given attribute values.
        """
        return self.portType(**kw)


    def listen(self, *a, **kw):
        """
        Pretend to bind a port.  Used as a stub implementation of a reactor
        listen method.  Subclasses should override and implement to append
        useful information to C{self.ports}.
        """
        raise NotImplementedError()


    def checkPort(self, port, alternatePort=None):
        """
        Assert that the given port has been properly created.

        @type port: L{DummyPort}
        @param port: A port which has been created by the code being tested.

        @type alternatePort: C{int}
        @param alternatePort: If not C{None}, the port number on which C{port}
        should be listening.
        """
        raise NotImplementedError()


    def setUp(self):
        self.filesdir = self.mktemp()
        self.store = Store(filesdir=self.filesdir)
        self.realFactory = ServerFactory()
        self.factory = DummyFactory(store=self.store, realFactory=self.realFactory)
        self.ports = []


    def test_portNumberAttribute(self):
        """
        Test that C{self.portType} remembers the port number it is told to
        listen on.
        """
        port = self.port(store=self.store, portNumber=self.lowPortNumber)
        self.assertEqual(port.portNumber, self.lowPortNumber)


    def test_interfaceAttribute(self):
        """
        Test that C{self.portType} remembers the interface it is told to listen
        on.
        """
        port = self.port(store=self.store, interface=self.someInterface)
        self.assertEqual(port.interface, self.someInterface)


    def test_factoryAttribute(self):
        """
        Test that C{self.portType} remembers the factory it is given to associate
        with its port.
        """
        port = self.port(store=self.store, factory=self.factory)
        self.assertIdentical(port.factory, self.factory)


    def test_service(self):
        """
        Test that C{self.portType} becomes a service on the store it is installed on.
        """
        port = self.port(store=self.store)
        installOn(port, self.store)

        self.assertEqual(
            list(self.store.powerupsFor(IService)),
            [port])


    def test_setServiceParent(self):
        """
        Test that the C{self.portType.setServiceParent} method adds the C{self.portType} to
        the Axiom Store Service as a child.
        """
        port = self.port(store=self.store)
        port.setServiceParent(self.store)
        self.failUnlessIn(port, list(IService(self.store)))


    def test_disownServiceParent(self):
        """
        Test that the C{self.portType.disownServiceParent} method removes the
        C{self.portType} from the Axiom Store Service.
        """
        port = self.port(store=self.store)
        port.setServiceParent(self.store)
        port.disownServiceParent()
        self.failIfIn(port, list(IService(self.store)))


    def test_serviceParent(self):
        """
        Test that C{self.portType} is a child of the store service after it is
        installed.
        """
        port = self.port(store=self.store)
        installOn(port, self.store)

        service = IServiceCollection(self.store)
        self.failUnlessIn(port, list(service))


    def _start(self, portNumber, methodName):
        port = self.port(store=self.store, portNumber=portNumber, factory=self.factory)
        port._listen = self.listen
        getattr(port, methodName)()
        return self.ports


    def _privilegedStartService(self, portNumber):
        return self._start(portNumber, 'privilegedStartService')


    def _startService(self, portNumber):
        return self._start(portNumber, 'startService')


    def test_startPrivilegedService(self):
        """
        Test that C{self.portType} binds a low-numbered port with the reactor when it
        is started with privilege.
        """
        ports = self._privilegedStartService(self.lowPortNumber)
        self.assertEqual(len(ports), 1)
        self.checkPort(ports[0])


    def test_dontStartPrivilegedService(self):
        """
        Test that C{self.portType} doesn't bind a high-numbered port with the
        reactor when it is started with privilege.
        """
        ports = self._privilegedStartService(self.highPortNumber)
        self.assertEqual(ports, [])


    def test_startServiceLow(self):
        """
        Test that C{self.portType} binds a low-numbered port with the reactor
        when it is started without privilege.
        """
        ports = self._startService(self.lowPortNumber)
        self.assertEqual(len(ports), 1)
        self.checkPort(ports[0])


    def test_startServiceHigh(self):
        """
        Test that C{self.portType} binds a high-numbered port with the reactor
        when it is started without privilege.
        """
        ports = self._startService(self.highPortNumber)
        self.assertEqual(len(ports), 1)
        self.checkPort(ports[0], self.highPortNumber)


    def test_startServiceNoInterface(self):
        """
        Test that C{self.portType} binds to all interfaces if no interface is
        explicitly specified.
        """
        port = self.port(store=self.store, portNumber=self.highPortNumber, factory=self.factory)
        port._listen = self.listen
        port.startService()
        self.assertEqual(self.ports[0].interface, '')


    def test_startServiceInterface(self):
        """
        Test that C{self.portType} binds to only the specified interface when
        instructed to.
        """
        port = self.port(store=self.store, portNumber=self.highPortNumber, factory=self.factory, interface=self.someInterface)
        port._listen = self.listen
        port.startService()
        self.assertEqual(self.ports[0].interface, self.someInterface)


    def test_startedOnce(self):
        """
        Test that C{self.portType} only binds one network port when
        C{privilegedStartService} and C{startService} are both called.
        """
        port = self.port(store=self.store, portNumber=self.lowPortNumber, factory=self.factory)
        port._listen = self.listen
        port.privilegedStartService()
        self.assertEqual(len(self.ports), 1)
        self.checkPort(self.ports[0])
        port.startService()
        self.assertEqual(len(self.ports), 1)


    def test_stopService(self):
        """
        Test that C{self.portType} cleans up its listening port when it is stopped.
        """
        port = self.port(store=self.store, portNumber=self.lowPortNumber, factory=self.factory)
        port._listen = self.listen
        port.startService()
        stopped = port.stopService()
        stopping = self.ports[0].stopping
        self.failIfIdentical(stopping, None)
        self.assertIdentical(stopped, stopping)


    def test_deletedFactory(self):
        """
        Test that the deletion of a C{self.portType}'s factory item results in the
        C{self.portType} being deleted.
        """
        port = self.port(store=self.store, portNumber=self.lowPortNumber, factory=self.factory)
        self.factory.deleteFromStore()
        self.assertEqual(list(self.store.query(self.portType)), [])


    def test_deletionDisownsParent(self):
        """
        Test that a deleted C{self.portType} no longer shows up in the children list
        of the service which used to be its parent.
        """
        port = self.port(store=self.store, portNumber=self.lowPortNumber, factory=self.factory)
        port.setServiceParent(self.store)
        port.deleteFromStore()
        service = IServiceCollection(self.store)
        self.failIfIn(port, list(service))



class TCPPortTests(PortTestsMixin, TestCase):
    """
    Tests for L{xmantissa.port.TCPPort}.
    """
    portType = TCPPort


    def checkPort(self, port, alternatePort=None):
        if alternatePort is None:
            alternatePort = self.lowPortNumber
        self.assertEqual(port.portNumber, alternatePort)
        self.assertEqual(port.factory, self.realFactory)


    def listen(self, port, factory, interface=''):
        self.ports.append(DummyPort(port, factory, interface=interface))
        return self.ports[-1]



class SSLPortTests(PortTestsMixin, TestCase):
    """
    Tests for L{xmantissa.port.SSLPort}.
    """
    portType = SSLPort

    def checkPort(self, port, alternatePort=None):
        if alternatePort is None:
            alternatePort = self.lowPortNumber
        self.assertEqual(port.portNumber, alternatePort)
        self.assertEqual(port.factory, self.realFactory)
        self.failUnless(isinstance(port.contextFactory, CertificateOptions))


    def port(self, certificatePath=None, **kw):
        if certificatePath is None:
            certificatePath = self.store.newFilePath('certificate.pem')
            assert not certificatePath.exists()
            certificatePath.setContent(CERTIFICATE_DATA + PRIVATEKEY_DATA)
        return self.portType(certificatePath=certificatePath, **kw)


    def listen(self, port, factory, contextFactory, interface=''):
        self.ports.append(DummyPort(port, factory, contextFactory, interface=interface))
        return self.ports[-1]


    def test_certificatePathAttribute(self):
        """
        Test that L{SSLPort} remembers the certificate filename it is given.
        """
        certificatePath = self.store.newFilePath('foo', 'bar')
        port = self.port(store=self.store, certificatePath=certificatePath)
        self.assertEqual(port.certificatePath, certificatePath)



class PortConfigurationCommandTests(TestCase):
    """
    Tests for the I{axiomatic port} command.
    """
    def setUp(self):
        """
        Override C{sys.stdout} to capture anything written by the port
        subcommand.
        """
        self.oldColumns = os.environ.get('COLUMNS')
        os.environ['COLUMNS'] = '80'
        self.stdout = sys.stdout
        sys.stdout = StringIO()


    def tearDown(self):
        """
        Restore the original value of C{sys.stdout}.
        """
        sys.stdout = self.stdout
        if self.oldColumns is not None:
            os.environ['COLUMNS'] = self.oldColumns

    def _makeConfig(self, store):
        """
        Create a L{PortConfiguration} instance with a properly set C{parent}
        attribute.
        """
        config = PortConfiguration()
        config.parent = CommandStub(store, "port")
        return config


    def assertSuccessStatus(self, options, arguments):
        """
        Parse the given arguments with the given options object and assert that
        L{SystemExit} is raised with an exit code of C{0}.
        """
        self.assertFailStatus(0, options, arguments)


    def assertFailStatus(self, code, options, arguments):
        """
        Parse the given arguments with the given options object and assert that
        L{SystemExit} is raised with the specified exit code.
        """
        exc = self.assertRaises(SystemExit, options.parseOptions, arguments)
        self.assertEqual(exc.args, (code,))


    def test_providesCommandInterface(self):
        """
        L{PortConfiguration} provides L{IAxiomaticCommand}.
        """
        verifyObject(IAxiomaticCommand, PortConfiguration)


    def test_axiomaticSubcommand(self):
        """
        L{PortConfiguration} is available as a subcommand of I{axiomatic}.
        """
        subCommands = AxiomaticOptions().subCommands
        [options] = [cmd[2] for cmd in subCommands if cmd[0] == 'port']
        self.assertIdentical(options, PortConfiguration)


    _portHelpText = (
        # This is a bit unfortunate.  It depends on what Options in Twisted
        # decides to spit out.  Note particularly the seemingly random amount
        # of trailing whitespace included on some lines.  The intent of tests
        # using this isn't really to ensure byte-identical results, but simply
        # to verify that help text is going to be shown to a user.  -exarkun
        "Usage: axiomatic [options] port [options]\n"
        "Options:\n"
        "      --version  \n"
        "      --help     Display this help and exit.\n"
        "\n"
        "This command allows for the inspection and modification of the "
        "configuration of\n"
        "network services in an Axiom store.\n"
        "Commands:\n"
        "    list        Show existing ports and factories.\n"
        "    delete      Delete existing ports.\n"
        "    create      Create new ports.\n"
        "\n")

    def test_implicitPortHelp(self):
        """
        When I{axiomatic port} is invoked with no arguments, usage information
        is written to standard out and the process exits successfully.
        """
        self.assertSuccessStatus(self._makeConfig(None), [])
        self.assertEqual(self._portHelpText, sys.stdout.getvalue())


    def test_explicitPortHelp(self):
        """
        When I{axiomatic port} is invoked with I{--help}, usage information is
        written to standard out.
        """
        self.assertSuccessStatus(self._makeConfig(None), ["--help"])
        self.assertEqual(self._portHelpText, sys.stdout.getvalue())


    _listHelpText = (
        "Usage: axiomatic [options] port [options] list [options]\n"
        "Options:\n"
        "      --version  \n"
        "      --help     Display this help and exit.\n"
        "\n"
        "Show the port/factory bindings in an Axiom store.\n"
        "\n")

    def test_explicitListHelp(self):
        """
        When I{axiomatic port list} is invoked with I{--help}, usage
        information for the C{list} subcommand is written to standard out.
        """
        self.assertSuccessStatus(self._makeConfig(None), ["list", "--help"])
        self.assertEqual(self._listHelpText, sys.stdout.getvalue())


    def test_listEmpty(self):
        """
        When I{axiomatic port list} is invoked, the ports which are currently
        configured in the system are displayed.
        """
        store = Store()
        self.assertSuccessStatus(self._makeConfig(store), ["list"])
        self.assertIn("There are no ports configured.", sys.stdout.getvalue())


    def test_listTCPPort(self):
        """
        When I{axiomatic port list} is invoked for a L{Store} which has a
        L{TCPPort} in it, the details of that port, including its factory, are
        written to stdout.
        """
        store = Store()
        factory = DummyFactory(store=store)
        port = TCPPort(
            store=store, factory=factory, portNumber=1234, interface=u"foo")
        self.assertSuccessStatus(self._makeConfig(store), ["list"])
        self.assertEqual(
            "%d) %r listening on:\n" % (factory.storeID, factory) +
            "  %d) TCP, interface %s, port %d\n" % (
                port.storeID, port.interface, port.portNumber),
            sys.stdout.getvalue())


    def test_listSSLPort(self):
        """
        When I{axiomatic port list} is invoked for a L{Store} which has an
        L{SSLPort} in it, the details of that port, including its factory, are
        written to stdout.
        """
        store = Store(filesdir=self.mktemp())
        factory = DummyFactory(store=store)
        port = SSLPort(
            store=store, factory=factory, portNumber=1234, interface=u"foo",
            certificatePath=store.filesdir.child("bar"))
        self.assertSuccessStatus(self._makeConfig(store), ["list"])
        self.assertEqual(
            "%d) %r listening on:\n" % (factory.storeID, factory) +
            "  %d) SSL, interface %s, port %d, certificate %s\n" % (
                port.storeID, port.interface, port.portNumber,
                port.certificatePath.path),
            sys.stdout.getvalue())


    def test_listAnyInterface(self):
        """
        I{axiomatic port list} displays a special string for a port bound to
        C{INADDR_ANY}.
        """
        store = Store()
        factory = DummyFactory(store=store)
        port = TCPPort(
            store=store, factory=factory, portNumber=1234, interface=u"")
        self.assertSuccessStatus(self._makeConfig(store), ["list"])
        self.assertEqual(
            "%d) %r listening on:\n" % (factory.storeID, factory) +
            "  %d) TCP, any interface, port %d\n" % (port.storeID, port.portNumber),
            sys.stdout.getvalue())


    def test_listSSLPortWithoutAttributes(self):
        """
        If there is an L{SSLPort} with no certificate or no port number (a
        rather invalid configuration), I{axiomatic port list} should show this
        in its output without producing an error.
        """
        store = Store()
        factory = DummyFactory(store=store)
        port = SSLPort(store=store, factory=factory)
        self.assertSuccessStatus(self._makeConfig(store), ["list"])
        self.assertEqual(
            "%d) %r listening on:\n" % (factory.storeID, factory) +
            "  %d) SSL, any interface, NO PORT, NO CERTIFICATE\n" % (
                port.storeID,),
            sys.stdout.getvalue())


    def test_listTwoPorts(self):
        """
        I{axiomatic port list} displays two different ports bound to the same
        factory together beneath that factory.
        """
        store = Store()
        factory = DummyFactory(store=store)
        portOne = TCPPort(
            store=store, factory=factory, portNumber=1234, interface=u"foo")
        portTwo = TCPPort(
            store=store, factory=factory, portNumber=2345, interface=u"bar")
        self.assertSuccessStatus(self._makeConfig(store), ["list"])
        self.assertEqual(
            "%d) %r listening on:\n" % (factory.storeID, factory) +
            "  %d) TCP, interface %s, port %d\n" % (
                portOne.storeID, portOne.interface, portOne.portNumber) +
            "  %d) TCP, interface %s, port %d\n" % (
                portTwo.storeID, portTwo.interface, portTwo.portNumber),
            sys.stdout.getvalue())


    def test_listTwoFactories(self):
        """
        I{axiomatic port list} displays two different factories separately from
        each other.
        """
        store = Store()
        factoryOne = DummyFactory(store=store)
        factoryTwo = DummyFactory(store=store)
        portOne = TCPPort(
            store=store, factory=factoryOne, portNumber=10, interface=u"foo")
        portTwo = TCPPort(
            store=store, factory=factoryTwo, portNumber=20, interface=u"bar")
        self.assertSuccessStatus(self._makeConfig(store), ["list"])
        self.assertEqual(
            "%d) %r listening on:\n" % (factoryOne.storeID, factoryOne) +
            "  %d) TCP, interface %s, port %d\n" % (
                portOne.storeID, portOne.interface, portOne.portNumber) +
            "%d) %r listening on:\n" % (factoryTwo.storeID, factoryTwo) +
            "  %d) TCP, interface %s, port %d\n" % (
                portTwo.storeID, portTwo.interface, portTwo.portNumber),
            sys.stdout.getvalue())


    def test_listUnlisteningFactory(self):
        """
        I{axiomatic port list} displays factories even if they aren't associate
        with any port.
        """
        store = Store()
        factory = DummyFactory(store=store)
        store.powerUp(factory, IProtocolFactoryFactory)
        self.assertSuccessStatus(self._makeConfig(store), ["list"])
        self.assertEqual(
            "%d) %r is not listening.\n" % (factory.storeID, factory),
            sys.stdout.getvalue())


    _deleteHelpText = (
        "Usage: axiomatic [options] port [options] delete [options]\n"
        "Options:\n"
        "      --port-identifier=  Identify a port for deletion.\n"
        "      --version           \n"
        "      --help              Display this help and exit.\n"
        "\n"
        "Delete an existing port binding from a factory. If a server is "
        "currently running\n"
        "using the database from which the port is deleted, the factory "
        "will *not* stop\n"
        "listening on that port until the server is restarted.\n"
        "\n")

    def test_explicitDeleteHelp(self):
        """
        If I{axiomatic port delete} is invoked with I{--help}, usage
        information for the C{delete} subcommand is written to standard out.
        """
        store = Store()
        self.assertSuccessStatus(self._makeConfig(store), ["delete", "--help"])
        self.assertEqual(self._deleteHelpText, sys.stdout.getvalue())


    def test_implicitDeleteHelp(self):
        """
        If I{axiomatic port delete} is invoked with no arguments, usage
        information for the C{delete} subcommand is written to standard out.
        """
        store = Store()
        self.assertSuccessStatus(self._makeConfig(store), ["delete"])
        self.assertEqual(self._deleteHelpText, sys.stdout.getvalue())


    def test_deletePorts(self):
        """
        I{axiomatic port delete} deletes each ports with a C{storeID} which is
        specified.
        """
        store = Store(filesdir=self.mktemp())
        factory = DummyFactory(store=store)
        deleteTCP = TCPPort(
            store=store, factory=factory, portNumber=10, interface=u"foo")
        keepTCP = TCPPort(
            store=store, factory=factory, portNumber=10, interface=u"bar")
        deleteSSL = SSLPort(
            store=store, factory=factory, portNumber=10, interface=u"baz",
            certificatePath=store.filesdir.child("baz"))
        keepSSL = SSLPort(
            store=store, factory=factory, portNumber=10, interface=u"quux",
            certificatePath=store.filesdir.child("quux"))
        self.assertSuccessStatus(
            self._makeConfig(store),
            ["delete",
             "--port-identifier", str(deleteTCP.storeID),
             "--port-identifier", str(deleteSSL.storeID)])
        self.assertEqual("Deleted.\n", sys.stdout.getvalue())
        self.assertEqual(list(store.query(TCPPort)), [keepTCP])
        self.assertEqual(list(store.query(SSLPort)), [keepSSL])


    def test_cannotDeleteOtherStuff(self):
        """
        I{axiomatic port delete} will not delete something which is neither a
        L{TCPPort} nor an L{SSLPort} and does not delete anything if an invalid
        port identifier is present in the command.
        """
        store = Store()
        factory = DummyFactory(store=store)
        tcp = TCPPort(
            store=store, factory=factory, interface=u"foo", portNumber=1234)
        self.assertFailStatus(
            1,
            self._makeConfig(store),
            ["delete",
             "--port-identifier", str(tcp.storeID),
             "--port-identifier", str(factory.storeID)])
        self.assertEqual(
            "%d does not identify a port.\n" % (factory.storeID,),
            sys.stdout.getvalue())
        self.assertEqual(list(store.query(DummyFactory)), [factory])
        self.assertEqual(list(store.query(TCPPort)), [tcp])


    def test_cannotDeleteNonExistent(self):
        """
        I{axiomatic port delete} writes a short error to standard output when a
        port-identifier is specified for which there is no corresponding store
        ID.
        """
        store = Store()
        self.assertFailStatus(
            1, self._makeConfig(store),
            ["delete", "--port-identifier", "12345"])
        self.assertEqual(
            "12345 does not identify an item.\n",
            sys.stdout.getvalue())


    _createHelpText = (
        "Usage: axiomatic [options] port [options] create [options]\n"
        "Options:\n"
        "      --strport=             A Twisted strports description of a "
        "port to add.\n"
        "      --factory-identifier=  Identifier for a protocol factory to "
        "associate with\n"
        "                             the new port.\n"
        "      --version              \n"
        "      --help                 Display this help and exit.\n"
        "\n"
        "Create a new port binding for an existing factory. If a server is "
        "currently\n"
        "running using the database in which the port is created, the "
        "factory will *not*\n"
        "be started on that port until the server is restarted.\n"
        "\n")

    def test_createImplicitHelp(self):
        """
        If I{axiomatic port create} is invoked with no arguments, usage
        information for the C{create} subcommand is written to standard out.
        """
        self.assertSuccessStatus(self._makeConfig(None), ["create"])
        self.assertEqual(self._createHelpText, sys.stdout.getvalue())


    def test_createExplicitHelp(self):
        """
        If I{axiomatic port create} is invoked with C{--help} as an argument,
        usage information for the C{add} subcommand is written to standard out.
        """
        self.assertSuccessStatus(self._makeConfig(None), ["create", "--help"])
        self.assertEqual(self._createHelpText, sys.stdout.getvalue())


    def test_createInvalidPortDescription(self):
        """
        If an invalid string is given for the C{strport} option of I{axiomatic
        port create}, a short error is written to standard output.
        """
        store = Store()
        factory = DummyFactory(store=store)
        self.assertFailStatus(
            1, self._makeConfig(store),
            ["create", "--strport", "xyz",
             "--factory-identifier", str(factory.storeID)])
        self.assertEqual(
            "'xyz' is not a valid port description.\n", sys.stdout.getvalue())


    def test_createNonExistentFactoryIdentifier(self):
        """
        If a storeID which is not associated with any item is given for the
        C{factory-identifier} option of I{axiomatic port create}, a short error
        is written to standard output.
        """
        store = Store()
        self.assertFailStatus(
            1, self._makeConfig(store),
            ["create", "--strport", "tcp:8080",
             "--factory-identifier", "123"])
        self.assertEqual(
            "123 does not identify an item.\n", sys.stdout.getvalue())


    def test_createNonFactoryIdentifier(self):
        """
        If a storeID which is associated with an item which does not provide
        L{IProtocolFactoryFactory} is given for the C{factory-identifier}
        option of I{axiomatic port create}, a short error is written to
        standard output.
        """
        store = Store()
        storeID = TCPPort(store=store).storeID
        self.assertFailStatus(
            1, self._makeConfig(store),
            ["create", "--strport", "tcp:8080",
             "--factory-identifier", str(storeID)])
        self.assertEqual(
            "%d does not identify a factory.\n" % (storeID,),
            sys.stdout.getvalue())


    def test_createTCPPort(self):
        """
        If given a valid strport description of a TCP port and the storeID of
        an extant factory, I{axiomatic port create} creates a new L{TCPPort}
        with the specified configuration and referring to that factory.  The
        port is also powered up on the store for L{IService}.
        """
        store = Store()
        factory = DummyFactory(store=store)
        self.assertSuccessStatus(
            self._makeConfig(store),
            ["create", "--strport", "tcp:8080",
             "--factory-identifier", str(factory.storeID)])
        self.assertEqual("Created.\n", sys.stdout.getvalue())
        [tcp] = list(store.query(TCPPort))
        self.assertEqual(tcp.portNumber, 8080)
        self.assertIdentical(tcp.factory, factory)
        self.assertEqual(list(store.interfacesFor(tcp)), [IService])


    def test_createSSLPortInvalidCertificate(self):
        """
        If given a certificate file which does not contain a PEM-format
        certificate, I{axiomatic port create} writes a short error message to
        standard output.
        """
        pemPath = FilePath(self.mktemp())
        pemPath.setContent("pem goes here")
        store = Store()
        factory = DummyFactory(store=store)
        self.assertFailStatus(
            1, self._makeConfig(store),
            ["create", "--strport", "ssl:8443:certKey=" + pemPath.path,
             "--factory-identifier", str(factory.storeID)])
        self.assertEqual(
            "Certificate file must use PEM format.\n",
            sys.stdout.getvalue())
        self.assertEqual(store.query(SSLPort).count(), 0)


    def test_createSSLPortNonExistentCertificateFile(self):
        """
        Specifying a certificate file which does not exist when creating an SSL
        port with I{axiomatic port create} causes a short error message to be
        written to stdout.
        """
        pemPath = FilePath(self.mktemp())
        pemPath.setContent(PRIVATEKEY_DATA)

        store = Store()
        factory = DummyFactory(store=store)
        self.assertFailStatus(
            1, self._makeConfig(store),
            ["create",
             "--strport", "ssl:8443:certKey=quux:privateKey=" + pemPath.path,
             "--factory-identifier", str(factory.storeID)])
        self.assertEqual(
            "Specified certificate file does not exist.\n",
            sys.stdout.getvalue())
        self.assertEqual(store.query(SSLPort).count(), 0)


    def test_createSSLPortNonExistentKeyFile(self):
        """
        Specifying a private key file which does not exist when creating an SSL
        port with I{axiomatic port create} causes a short error message to be
        written to stdout.
        """
        pemPath = FilePath(self.mktemp())
        pemPath.setContent(CERTIFICATE_DATA)

        store = Store()
        factory = DummyFactory(store=store)
        self.assertFailStatus(
            1, self._makeConfig(store),
            ["create",
             "--strport", "ssl:8443:privateKey=quux:certKey=" + pemPath.path,
             "--factory-identifier", str(factory.storeID)])
        self.assertEqual(
            "Specified private key file does not exist.\n",
            sys.stdout.getvalue())
        self.assertEqual(store.query(SSLPort).count(), 0)


    def test_createSSLPortInconsistentCertificateAndKeyFiles(self):
        """
        If different values are specified for the certificate file and the
        private key file when creating an SSL port with I{axiomatic port
        create}, a short error message is written to standard output.

        This reflects an implementation limitation which may be lifted in the
        future.
        """
        certPath = FilePath(self.mktemp())
        certPath.setContent(CERTIFICATE_DATA)
        keyPath = FilePath(self.mktemp())
        keyPath.setContent(PRIVATEKEY_DATA)

        store = Store()
        factory = DummyFactory(store=store)
        self.assertFailStatus(
            1, self._makeConfig(store),
            ["create",
             "--strport", "ssl:8443:privateKey=" + keyPath.path +
             ":certKey=" + certPath.path,
             "--factory-identifier", str(factory.storeID)])
        self.assertEqual(
            "You must specify the same file for certKey and privateKey.\n",
            sys.stdout.getvalue())
        self.assertEqual(store.query(SSLPort).count(), 0)


    def test_createSSLPortWithMethod(self):
        """
        SSL method configuration is unsupported and when I{axiomatic port
        create} is used to create an SSL port, an error is written to standard
        out if an attempt is made to specify a method.
        """
        pemPath = FilePath(self.mktemp())
        pemPath.setContent(CERTIFICATE_DATA + PRIVATEKEY_DATA)

        store = Store()
        factory = DummyFactory(store=store)
        self.assertFailStatus(
            1, self._makeConfig(store),
            ["create",
             "--strport", "ssl:8443:privateKey=" + pemPath.path +
             ":certKey=" + pemPath.path + ":sslmethod=TLSv1_METHOD",
             "--factory-identifier", str(factory.storeID)])
        self.assertEqual(
            "Only SSLv23_METHOD is supported.\n",
            sys.stdout.getvalue())
        self.assertEqual(store.query(SSLPort).count(), 0)


    def test_createSSLPort(self):
        """
        If a given valid strport description of an SSL port and the storeID of
        an extant factory, I{axiomatic port create} creates a new L{SSLPort}
        with the specified configuration and referring to that factory.  The
        certificate file specified is copied to a path inside the Store's files
        directory.  The port is also powered up on the store for L{IService}.
        """
        pemPath = FilePath(self.mktemp())
        pemPath.setContent(CERTIFICATE_DATA + PRIVATEKEY_DATA)
        store = Store(filesdir=self.mktemp())
        factory = DummyFactory(store=store)
        self.assertSuccessStatus(
            self._makeConfig(store),
            ["create", "--strport",
             "ssl:8443:certKey=" + pemPath.path +
             ":privateKey=" + pemPath.path,
             "--factory-identifier", str(factory.storeID)])
        self.assertEqual("Created.\n", sys.stdout.getvalue())
        [ssl] = list(store.query(SSLPort))
        self.assertEqual(ssl.portNumber, 8443)
        self.assertEqual(
            ssl.certificatePath.getContent(),
            CERTIFICATE_DATA + PRIVATEKEY_DATA)
        self.assertIdentical(ssl.factory, factory)
        self.assertEqual(
            pemPath.getContent(), CERTIFICATE_DATA + PRIVATEKEY_DATA)
        self.assertEqual(list(store.interfacesFor(ssl)), [IService])


    def test_createUnrecognized(self):
        """
        If given a strport description of an unrecognized port type,
        I{axiomatic port create} writes an error message to standard output.
        """
        store = Store()
        factory = DummyFactory(store=store)
        self.assertFailStatus(
            1, self._makeConfig(store),
            ["create", "--strport", "quux:foo",
             "--factory-identifier", str(factory.storeID)])
        self.assertEqual(
            "Unrecognized port type.\n", sys.stdout.getvalue())


    def test_createUnsupported(self):
        """
        If given a strport description of a valid type which is not supported
        by Mantissa, I{axiomatic port create} writes an error message to
        standard output.
        """
        store = Store()
        factory = DummyFactory(store=store)
        self.assertFailStatus(
            1, self._makeConfig(store),
            ["create", "--strport", "unix:/foo",
             "--factory-identifier", str(factory.storeID)])
        self.assertEqual(
            "Unsupported port type.\n", sys.stdout.getvalue())



from zope.interface import implements

from nevow import tags, athena, flat, loaders
from nevow.athena import expose

from formless.annotate import nameToLabel

from xmantissa.fragmentutils import PatternDictionary, dictFillSlots
from xmantissa import ixmantissa

# review the need to pass around instances of columns,
# rather than classes, same for actions also

class ColumnViewBase(object):
    def __init__(self, attributeID, displayName=None,
                 width=None, typeHint=None):
        """@param typeHint: text|datetime|action or None"""

        self.attributeID = attributeID
        if displayName is None:
            displayName = nameToLabel(attributeID)
        self.displayName = displayName
        self.width = width
        self.typeHint = typeHint

    def stanFromValue(self, idx, item, value):
        # called with the result of extractValue
        return unicode(value)

    def getWidth(self):
        if self.width is None:
            return ''
        else:
            return self.width

    def onclick(self, idx, item, value):
        return None

class EditableColumnView(ColumnViewBase):
    """Column View that can be edited within the table."""

    def __init__(self, attributeID, displayName=None, width=None,
                 typeHint=None, coercer=unicode, getter=None, setter=None,
                 **kw):
        super(EditableColumnView, self).__init__(attributeID=attributeID,
            displayName=displayName, width=width, typeHint=typeHint, **kw)
        if getter is not None:
            self.getter = getter
        if setter is not None:
            self.setter = setter

    def stanFromValue(self, idx, item, value):
        return EditableColumnField(self, idx, item, value)

    # custom getter/setter can be given as kwargs to override default

    def setter(self, item, value):
        setattr(item, self.attributeID, value)

    def getter(self, item):
        return getattr(item, self.attributeID)

class EditableColumnField(athena.LiveFragment):
    """Renders the input field in the table cell to edit data in-place."""

    docFactory = loaders.stan(
        tags.span(render=tags.directive('liveFragment'))[
            tags.span(render=tags.directive('field'))
        ])

    def __init__(self, original, idx, item, value, *a, **kw):
        athena.LiveFragment.__init__(self, original, *a, **kw)
        self.idx = idx
        self.item = item
        self.value = value

    def render_field(self, ctx, data):
        handler = "Nevow.Athena.Widget.get(this).callRemote('handleValueChange', this.value)"
        t = tags.input(type="text",
                onchange=handler,
                value=self.original.getter(self.item))
        return t

    def handleValueChange(self, value):
        return self.original.setter(self.item, self.original.coercer(value))
    expose(handleValueChange)



class DateColumnView(ColumnViewBase):
    def __init__(self, attributeID, displayName=None, width=None,
                 typeHint='datetime'):

        ColumnViewBase.__init__(self, attributeID, displayName,
                                width, typeHint)

    def stanFromValue(self, idx, item, value):
        # XXX timezones
        return value.asHumanly()

class ActionsColumnView(ColumnViewBase):
    def __init__(self, actions, width=None, typeHint='actions'):
        ColumnViewBase.__init__(self, 'Actions', width=width, typeHint=typeHint)
        self.actions = actions

    def stanFromValue(self, idx, item, value):
        # Value will generally be 'None' in this case...
        tag = tags.div()
        for action in self.actions:
            actionable = action.actionable(item)
            if actionable:
                iconURL = action.iconURL
            else:
                iconURL = action.disabledIconURL

            stan = tags.img(src=iconURL, **{'class' : 'tdb-action'})

            if actionable:
                linkstan = action.toLinkStan(idx, item)
                if linkstan is None:
                    handler = 'Mantissa.TDB.Controller.get(this).performAction(%r, %r); return false'
                    handler %= (action.actionID, idx)
                    stan = tags.a(href='#', onclick=handler)[stan]
                else:
                    stan = linkstan

            tag[stan]

        # at some point give the javascript the description to show
        # or something
        return tag

class Action(object):
    def __init__(self, actionID, iconURL, description, disabledIconURL=None):
        self.actionID = actionID
        self.iconURL = iconURL
        self.disabledIconURL = disabledIconURL
        self.description = description

    def performOn(self, item):
        """perform this action on the given item, returning
           None or a completion message that might be useful
           to the originator of the action"""
        raise NotImplementedError()

    def actionable(self, item):
        """return a boolean indicating whether it makes sense to
           perform this action on the given item"""
        raise NotImplementedError()

    def toLinkStan(self, idx, item):
        return None

class ToggleAction(Action):

    def actionable(self, item):
        return True

    def isOn(self, idx, item):
        raise NotImplementedError()

    def toLinkStan(self, idx, item):
        handler = 'Mantissa.TDB.Controller.get(this).performAction(%r, %r); return false'
        handler %= (self.actionID, idx)
        iconURL = (self.disabledIconURL, self.iconURL)[self.isOn(idx, item)]
        img = tags.img(src=iconURL, **{'class': 'tdb-action'})
        return tags.a(href='#', onclick=handler)[img]

class TabularDataView(athena.LiveFragment):
    implements(ixmantissa.INavigableFragment)

    jsClass = u'Mantissa.TDB.Controller'

    fragmentName = 'tdb'
    live = 'athena'
    title = ''
    patterns = None

    def __init__(self, model, columnViews, actions=(), width='', itemsCalled='Items'):
        """
        @param itemsCalled: something describing what a collection of tdb items is called,
                            e.g. 'Phone Numbers' or 'Clicks'
        """
        super(TabularDataView, self).__init__(model)
        self.columnViews = list(columnViews)
        if actions:
            self.columnViews.append(ActionsColumnView(actions))
        self.actions = {}
        for action in actions:
            self.actions[action.actionID] = action
        self.width = width
        self.itemsCalled = itemsCalled

    def constructTable(self):
        if self.patterns is None:
            self.patterns = PatternDictionary(self.docFactory)

        modelData = self.original.currentPage()

        if len(modelData) == 0:
            return self.patterns['no-rows'].fillSlots(
                                            'items-called', self.itemsCalled)

        tablePattern = self.patterns['table']

        headers = self.constructColumnHeaders()
        rows = self.constructRows(modelData)

        tablePattern = tablePattern.fillSlots('column-headers', list(headers))

        return tablePattern.fillSlots(
                'rows', list(rows)).fillSlots('width', self.width)

    def constructColumnHeaders(self):
        for cview in self.columnViews:
            model = self.original
            column = model.columns.get(cview.attributeID)

            if column is None:
                sortable = False
            else:
                sortable = column.sortAttribute() is not None

            if cview.attributeID == model.currentSortColumn.attributeID:
                headerPatternName = ['sorted-column-header-descending',
                                     'sorted-column-header-ascending'][model.isAscending]
            else:
                headerPatternName = ['column-header',
                                     'sortable-column-header'][sortable]

            header = self.patterns[headerPatternName].fillSlots(
                        'name', cview.displayName).fillSlots(
                                'width', cview.getWidth())

            if sortable:
                header = header.fillSlots('onclick',
                            'Mantissa.TDB.Controller.get(this).clickSort("%s")'%
                            (cview.attributeID,))

            yield header

    def constructRows(self, modelData):
        rowPattern = self.patterns['row']
        cellPattern = self.patterns['cell']

        for idx, row in enumerate(modelData):
            cells = []
            for cview in self.columnViews:
                value = row.get(cview.attributeID)
                cellContents = cview.stanFromValue(
                        idx, row['__item__'], value)
                if hasattr(cellContents, 'setFragmentParent'):
                    cellContents.setFragmentParent(self)
                handler = cview.onclick(idx, row['__item__'], value)
                cellStan = dictFillSlots(cellPattern,
                                         {'value': cellContents,
                                          'onclick': handler,
                                          'class': cview.typeHint})

                cells.append(cellStan)

            yield dictFillSlots(rowPattern,
                                {'cells': cells,
                                 'class': 'tdb-row-%s' % (idx,)})

    def getInitialArguments(self):
        return self.getPageState()

    def render_navigation(self, ctx, data):
        patterns = PatternDictionary(self.docFactory)
        return ctx.tag[patterns['navigation']()]

    def render_actions(self, ctx, data):
        return '(Actions not yet implemented)'

    def render_table(self, ctx, data):
        return self.constructTable()

    def replaceTable(self):
        # XXX TODO: the flatten here is encoding/decoding like 4 times; this
        # could be a lot faster.
        return unicode(flat.flatten(self.constructTable()), 'utf-8'), self.getPageState()
    expose(replaceTable)

    def getPageState(self):
        tdm = self.original
        return (tdm.hasPrevPage(), tdm.hasNextPage(),
                tdm.pageNumber, tdm.itemsPerPage, tdm.totalItems)
    expose(getPageState)

    def nextPage(self):
        self.original.nextPage()
        return self.replaceTable()
    expose(nextPage)

    def prevPage(self):
        self.original.prevPage()
        return self.replaceTable()
    expose(prevPage)

    def firstPage(self):
        self.original.firstPage()
        return self.replaceTable()
    expose(firstPage)

    def lastPage(self):
        self.original.lastPage()
        return self.replaceTable()
    expose(lastPage)

    def itemFromTargetID(self, targetID):
        modelData = list(self.original.currentPage())
        return modelData[targetID]['__item__']

    def performAction(self, actionID, targetID):
        target = self.itemFromTargetID(int(targetID))
        action = self.actions[actionID]
        result = action.performOn(target)
        return result, self.replaceTable()
    expose(performAction)

    def clickSort(self, attributeID):
        if attributeID == self.original.currentSortColumn.attributeID:
            self.original.resort(attributeID, not self.original.isAscending)
        else:
            self.original.resort(attributeID)
        self.original.firstPage()
        return self.replaceTable()
    expose(clickSort)

    def head(self):
        return None



# -*- test-case-name: xmantissa.test.test_port -*-

"""
Network port features for Mantissa services.

Provided herein are L{IService} L{Item} classes which can be used to take care
of most of the work required to run a network server within a Mantissa server.

Framework code should define an L{Item} subclass which implements
L{xmantissa.ixmantissa.IProtocolFactoryFactory} as desired.  No direct
interaction with the reactor nor specification of port or other network
configuration is necessary in that subclass.  Port types from this module can
be directly instantiated or configuration can be left up to another tool which
operates on arbitrary ports and L{IProtocolFactoryFactory} powerups (for
example, the administrative powerup L{xmantissa.webadmin.PortConfiguration}).

For example, a finger service might be defined in this way::

    from fingerproject import FingerFactory

    from axiom.item import Item
    from axiom.attributes import integer

    from xmantissa.ixmantissa import IProtocolFactoryFactory

    class Finger(Item):
        '''
        A finger (RFC 1288) server.
        '''
        implements(IProtocolFactoryFactory)
        powerupInterfaces = (IProtocolFactoryFactory,)

        requestCount = integer(doc='''
        The number of finger requests which have been responded to, ever.
        ''')

        def getFactory(self):
            return FingerFactory(self)

All concerns related to binding ports can be disregarded.  Once this item has
been added to a site store, an administrator will have access to it and may
configure it to listen on one or more ports.
"""

from zope.interface import implements

try:
    from OpenSSL import SSL
except ImportError:
    SSL = None

from twisted.application.service import IService, IServiceCollection
from twisted.application.strports import parse
from twisted.internet.ssl import PrivateCertificate, CertificateOptions
from twisted.python.reflect import qual
from twisted.python.usage import Options
from twisted.python.filepath import FilePath

from axiom.item import Item, declareLegacyItem, normalize
from axiom.attributes import inmemory, integer, reference, path, text
from axiom.upgrade import registerAttributeCopyingUpgrader
from axiom.dependency import installOn
from axiom.scripts.axiomatic import AxiomaticCommand, AxiomaticSubCommand

from xmantissa.ixmantissa import IProtocolFactoryFactory



class PortMixin:
    """
    Mixin implementing most of L{IService} as would be appropriate for an Axiom
    L{Item} subclass in order to manage the lifetime of an
    L{twisted.internet.interfaces.IListeningPort}.
    """
    implements(IService)

    powerupInterfaces = (IService,)

    # Required by IService but unused by this code.
    name = None

    def activate(self):
        self.parent = None
        self._listen = None
        self.listeningPort = None


    def installed(self):
        """
        Callback invoked after this item has been installed on a store.

        This is used to set the service parent to the store's service object.
        """
        self.setServiceParent(self.store)


    def deleted(self):
        """
        Callback invoked after a transaction in which this item has been
        deleted is committed.

        This is used to remove this item from its service parent, if it has
        one.
        """
        if self.parent is not None:
            self.disownServiceParent()


    # IService
    def setServiceParent(self, parent):
        IServiceCollection(parent).addService(self)
        self.parent = parent


    def disownServiceParent(self):
        IServiceCollection(self.parent).removeService(self)
        self.parent = None


    def privilegedStartService(self):
        if self.portNumber < 1024:
            self.listeningPort = self.listen()


    def startService(self):
        if self.listeningPort is None:
            self.listeningPort = self.listen()


    def stopService(self):
        d = self.listeningPort.stopListening()
        self.listeningPort = None
        return d



class TCPPort(PortMixin, Item):
    """
    An Axiom Service Item which will bind a TCP port to a protocol factory when
    it is started.
    """
    schemaVersion = 2

    portNumber = integer(doc="""
    The TCP port number on which to listen.
    """)

    interface = text(doc="""
    The hostname to bind to.
    """, default=u'')

    factory = reference(doc="""
    An Item with a C{getFactory} method which returns a Twisted protocol
    factory.
    """, whenDeleted=reference.CASCADE)

    parent = inmemory(doc="""
    A reference to the parent service of this service, whenever there is a
    parent.
    """)

    _listen = inmemory(doc="""
    An optional reference to a callable implementing the same interface as
    L{IReactorTCP.listenTCP}.  If set, this will be used to bind a network
    port.  If not set, the reactor will be imported and its C{listenTCP} method
    will be used.
    """)

    listeningPort = inmemory(doc="""
    A reference to the L{IListeningPort} returned by C{self.listen} which is
    set whenever there there is one listening.
    """)

    def listen(self):
        if self._listen is not None:
            _listen = self._listen
        else:
            from twisted.internet import reactor
            _listen = reactor.listenTCP
        return _listen(self.portNumber, self.factory.getFactory(),
                       interface=self.interface.encode('ascii'))

declareLegacyItem(
    typeName=normalize(qual(TCPPort)),
    schemaVersion=1,
    attributes=dict(
        portNumber=integer(),
        factory=reference(),
        parent=inmemory(),
        _listen=inmemory(),
        listeningPort=inmemory()))

registerAttributeCopyingUpgrader(TCPPort, 1, 2)



class SSLPort(PortMixin, Item):
    """
    An Axiom Service Item which will bind a TCP port to a protocol factory when
    it is started.
    """
    schemaVersion = 2

    portNumber = integer(doc="""
    The TCP port number on which to listen.
    """)

    interface = text(doc="""
    The hostname to bind to.
    """, default=u'')

    certificatePath = path(doc="""
    Name of the file containing the SSL certificate to use for this server.
    """)

    factory = reference(doc="""
    An Item with a C{getFactory} method which returns a Twisted protocol
    factory.
    """, whenDeleted=reference.CASCADE)

    parent = inmemory(doc="""
    A reference to the parent service of this service, whenever there is a
    parent.
    """)

    _listen = inmemory(doc="""
    An optional reference to a callable implementing the same interface as
    L{IReactorTCP.listenTCP}.  If set, this will be used to bind a network
    port.  If not set, the reactor will be imported and its C{listenTCP} method
    will be used.
    """)

    listeningPort = inmemory(doc="""
    A reference to the L{IListeningPort} returned by C{self.listen} which is
    set whenever there there is one listening.
    """)


    def getContextFactory(self):
        if SSL is None:
            raise RuntimeError("No SSL support: you need to install OpenSSL.")
        cert = PrivateCertificate.loadPEM(
            self.certificatePath.open().read())
        certOpts = CertificateOptions(
            cert.privateKey.original,
            cert.original,
            requireCertificate=False,
            method=SSL.SSLv23_METHOD)
        return certOpts


    def listen(self):
        if self._listen is not None:
            _listen = self._listen
        else:
            from twisted.internet import reactor
            _listen = reactor.listenSSL
        return _listen(
            self.portNumber,
            self.factory.getFactory(),
            self.getContextFactory(),
            interface=self.interface.encode('ascii'))

declareLegacyItem(
    typeName=normalize(qual(SSLPort)),
    schemaVersion=1,
    attributes=dict(
        portNumber=integer(),
        certificatePath=path(),
        factory=reference(),
        parent=inmemory(),
        _listen=inmemory(),
        listeningPort=inmemory()))

registerAttributeCopyingUpgrader(SSLPort, 1, 2)


class ListOptions(Options):
    """
    I{axiomatic port} subcommand for displaying the ports which are currently
    set up in a store.
    """
    longdesc = "Show the port/factory bindings in an Axiom store."

    def postOptions(self):
        """
        Display details about the ports which already exist.
        """
        store = self.parent.parent.getStore()
        port = None
        factories = {}
        for portType in [TCPPort, SSLPort]:
            for port in store.query(portType):
                key = port.factory.storeID
                if key not in factories:
                    factories[key] = (port.factory, [])
                factories[key][1].append(port)
        for factory in store.powerupsFor(IProtocolFactoryFactory):
            key = factory.storeID
            if key not in factories:
                factories[key] = (factory, [])
        def key((factory, ports)):
            return factory.storeID
        for factory, ports in sorted(factories.values(), key=key):
            if ports:
                print '%d) %r listening on:' % (factory.storeID, factory)
                for port in ports:
                    if port.interface:
                        interface = "interface " + port.interface
                    else:
                        interface = "any interface"
                    if isinstance(port, TCPPort):
                        print '  %d) TCP, %s, port %d' % (
                            port.storeID, interface, port.portNumber)
                    else:
                        if port.certificatePath is not None:
                            pathPart = 'certificate %s' % (
                                port.certificatePath.path,)
                        else:
                            pathPart = 'NO CERTIFICATE'
                        if port.portNumber is not None:
                            portPart = 'port %d' % (port.portNumber,)
                        else:
                            portPart = 'NO PORT'
                        print '  %d) SSL, %s, %s, %s' % (
                            port.storeID, interface, portPart, pathPart)
            else:
                print '%d) %r is not listening.' % (factory.storeID, factory)
        if not factories:
            print "There are no ports configured."
        raise SystemExit(0)



class DeleteOptions(Options):
    """
    I{axiomatic port} subcommand for removing existing ports.

    @type portIdentifiers: C{list} of C{int}
    @ivar portIdentifiers: The store IDs of the ports to be deleted, built up
        by the I{port-identifier} parameter.
    """
    longdesc = (
        "Delete an existing port binding from a factory.  If a server is "
        "currently running using the database from which the port is deleted, "
        "the factory will *not* stop listening on that port until the server "
        "is restarted.")

    def __init__(self):
        Options.__init__(self)
        self.portIdentifiers = []


    def opt_port_identifier(self, storeID):
        """
        Identify a port for deletion.
        """
        self.portIdentifiers.append(int(storeID))


    def _delete(self, store, portIDs):
        """
        Try to delete the ports with the given store IDs.

        @param store: The Axiom store from which to delete items.

        @param portIDs: A list of Axiom store IDs for TCPPort or SSLPort items.

        @raise L{SystemExit}: If one of the store IDs does not identify a port
            item.
        """
        for portID in portIDs:
            try:
                port = store.getItemByID(portID)
            except KeyError:
                print "%d does not identify an item." % (portID,)
                raise SystemExit(1)
            if isinstance(port, (TCPPort, SSLPort)):
                port.deleteFromStore()
            else:
                print "%d does not identify a port." % (portID,)
                raise SystemExit(1)


    def postOptions(self):
        """
        Delete the ports specified with the port-identifier option.
        """
        if self.portIdentifiers:
            store = self.parent.parent.getStore()
            store.transact(self._delete, store, self.portIdentifiers)
            print "Deleted."
            raise SystemExit(0)
        else:
            self.opt_help()



class CreateOptions(AxiomaticSubCommand):
    """
    I{axiomatic port} subcommand for creating new ports.
    """
    name = "create"
    longdesc = (
        "Create a new port binding for an existing factory.  If a server is "
        "currently running using the database in which the port is created, "
        "the factory will *not* be started on that port until the server is "
        "restarted.")

    _pemFormatError = ('PEM routines', 'PEM_read_bio', 'no start line')
    _noSuchFileError = ('system library', 'fopen', 'No such file or directory')
    _certFileError = ('SSL routines', 'SSL_CTX_use_certificate_file', 'system lib')
    _keyFileError = ('SSL routines', 'SSL_CTX_use_PrivateKey_file', 'system lib')

    optParameters = [
        ("strport", None, None,
         "A Twisted strports description of a port to add."),
        ("factory-identifier", None, None,
         "Identifier for a protocol factory to associate with the new port.")]


    def postOptions(self):
        strport = self['strport']
        factoryIdentifier = self['factory-identifier']
        if strport is None or factoryIdentifier is None:
            self.opt_help()

        store = self.parent.parent.getStore()
        storeID = int(factoryIdentifier)
        try:
            factory = store.getItemByID(storeID)
        except KeyError:
            print "%d does not identify an item." % (storeID,)
            raise SystemExit(1)
        else:
            if not IProtocolFactoryFactory.providedBy(factory):
                print "%d does not identify a factory." % (storeID,)
                raise SystemExit(1)
            else:
                try:
                    kind, args, kwargs = parse(strport, factory)
                except ValueError:
                    print "%r is not a valid port description." % (strport,)
                    raise SystemExit(1)
                except KeyError:
                    print "Unrecognized port type."
                    raise SystemExit(1)
                except SSL.Error, e:
                    if self._pemFormatError in e.args[0]:
                        print 'Certificate file must use PEM format.'
                        raise SystemExit(1)
                    elif self._noSuchFileError in e.args[0]:
                        if self._certFileError in e.args[0]:
                            print "Specified certificate file does not exist."
                            raise SystemExit(1)
                        elif self._keyFileError in e.args[0]:
                            print "Specified private key file does not exist."
                            raise SystemExit(1)
                        else:
                            # Note, no test coverage.
                            raise
                    else:
                        # Note, no test coverage.
                        raise
                else:
                    try:
                        method = getattr(self, 'create_' + kind)
                    except AttributeError:
                        print "Unsupported port type."
                        raise SystemExit(1)
                    else:
                        port = method(store, *args, **kwargs)
                        installOn(port, store)
                        print "Created."
        raise SystemExit(0)


    def create_TCP(self, store, port, factory, backlog, interface):
        """
        Create a new L{TCPPort} with the specified parameters.
        """
        return TCPPort(
            store=store, portNumber=port,
            factory=factory, interface=self.decodeCommandLine(interface))


    def create_SSL(self, store, port, factory, context, backlog, interface):
        """
        Create a new L{SSLPort} with the specified parameters.

        @type context: L{DefaultOpenSSLContextFactory}
        """
        key = context.privateKeyFileName
        cert = context.certificateFileName
        if key != cert:
            print "You must specify the same file for certKey and privateKey."
            raise SystemExit(1)
        elif context.sslmethod != SSL.SSLv23_METHOD:
            print "Only SSLv23_METHOD is supported."
            raise SystemExit(1)
        else:
            port = SSLPort(
                store=store, portNumber=port, factory=factory,
                interface=self.decodeCommandLine(interface))
            targetDir = store.filesdir.child(str(port.storeID))
            targetDir.makedirs()
            targetPath = targetDir.child("cert.pem")
            port.certificatePath = targetPath
            FilePath(key).copyTo(targetPath)
            return port



class PortConfiguration(AxiomaticCommand):
    """
    Axiomatic subcommand plugin for inspecting and modifying port
    configuration.
    """
    subCommands = [("list", None, ListOptions, "Show existing ports and factories."),
                   ("delete", None, DeleteOptions, "Delete existing ports."),
                   ("create", None, CreateOptions, "Create new ports.")]

    name = "port"
    description = "Examine, create, and destroy network servers."
    longdesc = (
        "This command allows for the inspection and modification of the "
        "configuration of network services in an Axiom store.")

    def postOptions(self):
        """
        If nothing else happens, display usage information.
        """
        self.opt_help()



__all__ = ['TCPPort', 'SSLPort', 'PortConfiguration']



# Copyright 2007 Divmod, Inc.
# See LICENSE file for details

"""
Tests for L{xmantissa.webnav}.
"""

from twisted.trial import unittest

from epsilon.structlike import record

from axiom.store import Store
from axiom.dependency import installOn

from nevow.url import URL
from nevow import tags, context
from nevow.testutil import FakeRequest

from xmantissa import webnav
from xmantissa.webapp import PrivateApplication



class FakeNavigator(record('tabs')):
    def getTabs(self):
        return self.tabs


class NavConfigTests(unittest.TestCase):
    """
    Tests for free functions in L{xmantissa.webnav}.
    """
    def test_tabMerge(self):
        """
        L{webnav.getTabs} should combine tabs from the L{INavigableElement}
        providers passed to it into a single structure.  It should preserve the
        attributes of all of the tabs and order them and their children by
        priority.
        """
        nav = webnav.getTabs([
                FakeNavigator([webnav.Tab('Hello', 1, 0.5,
                                          [webnav.Tab('Super', 2, 1.0, (), False, '/Super/2'),
                                           webnav.Tab('Mega', 3, 0.5, (), False, '/Mega/3')],
                                          False, '/Hello/1')]),
                FakeNavigator([webnav.Tab('Hello', 4, 1.,
                                          [webnav.Tab('Ultra', 5, 0.75, (), False, '/Ultra/5'),
                                           webnav.Tab('Hyper', 6, 0.25, (), False, '/Hyper/6')],
                                          True, '/Hello/4'),
                               webnav.Tab('Goodbye', 7, 0.9, (), True, '/Goodbye/7')])])

        hello, goodbye = nav
        self.assertEqual(hello.name, 'Hello')
        self.assertEqual(hello.storeID, 4)
        self.assertEqual(hello.priority, 1.0)
        self.assertEqual(hello.authoritative,True)
        self.assertEqual(hello.linkURL, '/Hello/4')

        super, ultra, mega, hyper = hello.children
        self.assertEqual(super.name, 'Super')
        self.assertEqual(super.storeID, 2)
        self.assertEqual(super.priority, 1.0)
        self.assertEqual(super.authoritative, False)
        self.assertEqual(super.linkURL, '/Super/2')

        self.assertEqual(ultra.name, 'Ultra')
        self.assertEqual(ultra.storeID, 5)
        self.assertEqual(ultra.priority, 0.75)
        self.assertEqual(ultra.authoritative, False)
        self.assertEqual(ultra.linkURL, '/Ultra/5')

        self.assertEqual(mega.name, 'Mega')
        self.assertEqual(mega.storeID, 3)
        self.assertEqual(mega.priority, 0.5)
        self.assertEqual(mega.authoritative, False)
        self.assertEqual(mega.linkURL, '/Mega/3')

        self.assertEqual(hyper.name, 'Hyper')
        self.assertEqual(hyper.storeID, 6)
        self.assertEqual(hyper.priority, 0.25)
        self.assertEqual(hyper.authoritative, False)
        self.assertEqual(hyper.linkURL, '/Hyper/6')

        self.assertEqual(goodbye.name, 'Goodbye')
        self.assertEqual(goodbye.storeID, 7)
        self.assertEqual(goodbye.priority, 0.9)
        self.assertEqual(goodbye.authoritative, True)
        self.assertEqual(goodbye.linkURL, '/Goodbye/7')


    def test_setTabURLs(self):
        """
        Check that L{webnav.setTabURLs} correctly sets the C{linkURL}
        attribute of L{webnav.Tab} instances to the result of
        passing tab.storeID to L{xmantissa.ixmantissa.IWebTranslator.linkTo}
        if C{linkURL} is not set, and that it leaves it alone if it is
        """

        s = Store()

        privapp = PrivateApplication(store=s)
        installOn(privapp,s)

        tabs = [webnav.Tab('PrivateApplication', privapp.storeID, 0),
                webnav.Tab('Something Else', None, 0, linkURL='/foo/bar')]

        webnav.setTabURLs(tabs, privapp)

        self.assertEqual(tabs[0].linkURL, privapp.linkTo(privapp.storeID))
        self.assertEqual(tabs[1].linkURL, '/foo/bar')


    def test_getSelectedTabExactMatch(self):
        """
        Check that L{webnav.getSelectedTab} returns the tab whose C{linkURL}
        attribute exactly matches the path of the L{nevow.url.URL} it is passed
        """

        tabs = list(webnav.Tab(str(i), None, 0, linkURL='/' + str(i))
                        for i in xrange(5))

        for (i, tab) in enumerate(tabs):
            selected = webnav.getSelectedTab(tabs, URL.fromString(tab.linkURL))
            self.assertIdentical(selected, tab)

        selected = webnav.getSelectedTab(tabs, URL.fromString('/XYZ'))
        self.failIf(selected)


    def test_getSelectedTabPrefixMatch(self):
        """
        Check that L{webnav.getSelectedTab} returns the tab whose C{linkURL}
        attribute contains the longest prefix of path segments that appears
        at the beginning of the L{nevow.url.URL} it is passed (if there is not
        an exact match)
        """

        tabs = [webnav.Tab('thing1', None, 0, linkURL='/a/b/c/d'),
                webnav.Tab('thing2', None, 0, linkURL='/a/b/c')]

        def assertSelected(tab):
            selected = webnav.getSelectedTab(tabs, URL.fromString('/a/b/c/d/e'))
            self.assertIdentical(selected, tab)

        assertSelected(tabs[0])
        tabs.reverse()
        assertSelected(tabs[1])

        tabs.append(webnav.Tab('thing3', None, 0, linkURL='a/b/c/e/e'))
        assertSelected(tabs[1])

        t = webnav.Tab('thing4', None, 0, linkURL='/a/b/c/d/e')
        tabs.append(t)
        assertSelected(t)



class FakeTranslator(object):
    """
    A dumb translator which follows a very simple translation rule and can only
    translate in one direction.
    """
    def linkTo(self, obj):
        """
        Return a fake link based on the given object.
        """
        return '/link/' + str(obj)



class RendererTests(unittest.TestCase):
    """
    Tests for certain free functions in L{xmantissa.webnav} which render
    different things.
    """
    def test_startMenuSetsTabURLs(self):
        """
        L{Tabs<Tab>} which have C{None} for a C{linkURL} attribute should have
        a value set for that attribute based on the L{IWebTranslator} passed to
        L{startMenu}.
        """
        tab = webnav.Tab('alpha', 123, 0)
        webnav.startMenu(FakeTranslator(), [tab], tags.span())
        self.assertEqual(tab.linkURL, '/link/123')


    def test_startMenuRenders(self):
        """
        Test that the L{startMenu} renderer creates a tag for each tab, filling
        its I{href}, I{name}, and I{kids} slots.
        """
        tabs = [
            webnav.Tab('alpha', 123, 0),
            webnav.Tab('beta', 234, 0)]
        node = tags.span[tags.div(pattern='tab')]

        tag = webnav.startMenu(FakeTranslator(), tabs, node)
        self.assertEqual(tag.tagName, 'span')
        navTags = list(tag.slotData['tabs'])
        self.assertEqual(len(navTags), 2)
        alpha, beta = navTags
        self.assertEqual(alpha.slotData['name'], 'alpha')
        self.assertEqual(alpha.slotData['href'], '/link/123')
        self.assertEqual(alpha.slotData['kids'], '')
        self.assertEqual(beta.slotData['name'], 'beta')
        self.assertEqual(beta.slotData['href'], '/link/234')
        self.assertEqual(beta.slotData['kids'], '')


    def test_settingsLink(self):
        """
        L{settingsLink} should add a link to the settings item supplied as a
        child of the tag supplied.
        """
        self.storeID = 123
        node = tags.span()
        tag = webnav.settingsLink(FakeTranslator(), self, node)
        self.assertEqual(tag.tagName, 'span')
        self.assertEqual(tag.children, ['/link/123'])


    def _renderAppNav(self, tabs, template=None):
        """
        Render application navigation and return the resulting tag.

        @param template: a Tag containing a template for navigation.
        """
        if template is None:
            template = tags.span[
                tags.div(pattern='app-tab'),
                tags.div(pattern='tab-contents')]
        ctx = context.WebContext(tag=template)
        request = FakeRequest()
        ctx.remember(request)
        return webnav.applicationNavigation(ctx, FakeTranslator(), tabs)


    def test_applicationNavigation(self):
        """
        Test that the L{applicationNavigation} renderer creates a tag for each
        tab, fillings I{name} and I{tab-contents} slots.
        """
        tag = self._renderAppNav([
            webnav.Tab('alpha', 123, 0),
            webnav.Tab('beta', 234, 0)])
        self.assertEqual(tag.tagName, 'span')
        navTags = list(tag.slotData['tabs'])
        self.assertEqual(len(navTags), 2)
        alpha, beta = navTags
        self.assertEqual(alpha.slotData['name'], 'alpha')
        alphaContents = alpha.slotData['tab-contents']
        self.assertEqual(alphaContents.slotData['href'], '/link/123')
        self.assertEqual(beta.slotData['name'], 'beta')
        betaContents = beta.slotData['tab-contents']
        self.assertEqual(betaContents.slotData['href'], '/link/234')


    def test_applicationNavigationChildren(self):
        """
        The L{applicationNavigation} renderer should fill the 'subtabs' slot
        with copies of the 'subtab' pattern for each tab, if that pattern is
        present.  (This is only tested to one level of depth because we
        currently only support one level of depth.)
        """
        tag = self._renderAppNav(
            [webnav.Tab('alpha', 123, 0),
             webnav.Tab('beta', 234, 0, children=[
                        webnav.Tab('gamma', 345, 0),
                        webnav.Tab('delta', 456, 0)])],
            tags.span[tags.div(pattern='app-tab'),
                      tags.div(pattern='tab-contents'),
                      tags.div(pattern='subtab'),
                      tags.div(pattern='subtab-contents',
                               class_='subtab-contents-class')])
        navTags = list(tag.slotData['tabs'])
        self.assertEqual(len(navTags), 2)
        alpha, beta = navTags
        self.assertEqual(alpha.slotData['subtabs'], [])
        self.assertEqual(len(beta.slotData['subtabs']), 2)
        subtab1 = beta.slotData['subtabs'][0]
        self.assertEqual(subtab1.slotData['name'],
                         'gamma')
        self.assertEqual(subtab1.slotData['href'], '/link/345')
        self.assertEqual(subtab1.slotData['tab-contents'].attributes['class'],
                         'subtab-contents-class')
        subtab2 = beta.slotData['subtabs'][1]
        self.assertEqual(subtab2.slotData['name'],
                         'delta')
        self.assertEqual(subtab2.slotData['href'], '/link/456')
        self.assertEqual(subtab2.slotData['tab-contents'].attributes['class'],
                         'subtab-contents-class')


    def test_applicationNavigationMissingSubtabsPattern(self):
        """
        The L{applicationNavigation} renderer should fill the 'subtabs' slot
        with the empty list if the 'subtabs' pattern is not found.  This is to
        ensure that it remains compatible with older customized 'shell'
        templates.
        """
        tag = self._renderAppNav([
                webnav.Tab("alpha", 123, 0,
                           children=[webnav.Tab("beta", 234, 0)])])
        navTags = list(tag.slotData['tabs'])
        self.assertEqual(navTags[0].slotData['subtabs'], [])



# -*- test-case-name: xmantissa.test.test_terminal -*-
# Copyright 2009 Divmod, Inc.  See LICENSE file for details

"""
This module provides an extensible terminal server for Mantissa, exposed over
L{SSHv2<twisted.conch>}.  The server supports username/password authentication
and encrypted communication.  It can be extended by applications which provide
L{ITerminalServerFactory} powerups to create L{ITerminalProtocol} providers.
"""

from md5 import md5

from Crypto.PublicKey import RSA

from zope.interface import implements

from twisted.python.randbytes import secureRandom

from twisted.python.components import Componentized
from twisted.cred.portal import IRealm, Portal
from twisted.cred.checkers import ICredentialsChecker
from twisted.conch.interfaces import IConchUser, ISession
from twisted.conch.ssh.factory import SSHFactory
from twisted.conch.ssh.keys import Key
from twisted.conch.manhole_ssh import TerminalUser, TerminalSession, TerminalSessionTransport
from twisted.conch.insults.insults import ServerProtocol, TerminalProtocol
from twisted.conch.insults.window import TopWindow, VBox, Border, Button
from twisted.conch.manhole import ColoredManhole

from axiom.iaxiom import IPowerupIndirector
from axiom.item import Item
from axiom.attributes import bytes
from axiom.dependency import dependsOn
from axiom.userbase import LoginSystem, getAccountNames

from xmantissa.ixmantissa import IProtocolFactoryFactory, ITerminalServerFactory
from xmantissa.ixmantissa import IViewer
from xmantissa.sharing import getAccountRole


__metaclass__ = type


def _generate():
    """
    Generate a new SSH key pair.
    """
    key = RSA.generate(1024, secureRandom)
    return Key(key).toString('openssh')



class _AuthenticatedShellViewer:
    """
    L{_AuthenticatedShellViewer} is an L{IViewer} implementation used to
    indicate to shell applications which user is trying to access them.

    @ivar _accountNames: A list of account names associated with this viewer.
        This is in the same form as the return value of L{getAccountNames}.
    """
    implements(IViewer)

    def __init__(self, accountNames):
        self._accountNames = accountNames


    def roleIn(self, userStore):
        """
        Get the authenticated role for the user represented by this view in the
        given user store.
        """
        return getAccountRole(userStore, self._accountNames)



class SecureShellConfiguration(Item):
    """
    Configuration object for a Mantissa SSH server.
    """
    powerupInterfaces = (IProtocolFactoryFactory,)
    implements(*powerupInterfaces)

    loginSystem = dependsOn(LoginSystem)

    hostKey = bytes(
        doc="""
        An OpenSSH-format string giving the host key for this server.
        """, allowNone=False, defaultFactory=_generate)


    def __repr__(self):
        """
        Return a summarized representation of this item.
        """
        fmt = "SecureShellConfiguration(storeID=%d, hostKeyFingerprint='%s')"
        privateKey = Key.fromString(data=self.hostKey)
        publicKeyBlob = privateKey.blob()
        fingerprint = md5(publicKeyBlob).hexdigest()
        return fmt % (self.storeID, fingerprint)


    def getFactory(self):
        """
        Create an L{SSHFactory} which allows access to Mantissa accounts.
        """
        privateKey = Key.fromString(data=self.hostKey)
        public = privateKey.public()
        factory = SSHFactory()
        factory.publicKeys = {'ssh-rsa': public}
        factory.privateKeys = {'ssh-rsa': privateKey}
        factory.portal = Portal(
            IRealm(self.store), [ICredentialsChecker(self.store)])
        return factory



class _ReturnToMenuWrapper:
    """
    L{ITerminalTransport} wrapper which changes only the behavior of the
    C{loseConnection} method.  Rather than allowing the terminal to be
    disconnected, a L{ShellServer} is re-activated.

    @ivar _shell: The L{ShellServer} to re-activate when disconnection is
        attempted.

    @ivar _terminal: The L{ITerminalTransport} to which to proxy all attribute
        lookups other than C{loseConnection}.
    """
    def __init__(self, shell, terminal):
        self._shell = shell
        self._terminal = terminal


    def loseConnection(self):
        self._shell.reactivate()


    def __getattr__(self, name):
        return getattr(self._terminal, name)



class ShellServer(TerminalProtocol):
    """
    A terminal protocol which finds L{ITerminalServerFactory} powerups in the
    same store and presents the option of beginning a session with one of them.

    @ivar _store: The L{Store} which will be searched for
        L{ITerminalServerFactory} powerups.

    @ivar _protocol: If an L{ITerminalServerFactory} has been selected to
        interact with, then this attribute refers to the L{ITerminalProtocol}
        produced by that factory's C{buildTerminalProtocol} method.  Input from
        the terminal is delivered to this protocol.  This attribute is C{None}
        whenever the "main menu" user interface is being displayed.

    @ivar _window: A L{TopWindow} instance which contains the "main menu" user
        interface.  Whenever the C{_protocol} attribute is C{None}, input is
        directed to this object instead.  Whenever the C{_protocol} attribute
        is not C{None}, this window is hidden.
    """
    _width = 80
    _height = 24

    _protocol = None

    def __init__(self, store):
        TerminalProtocol.__init__(self)
        self._store = store


    def _draw(self):
        """
        Call the drawing API for the main menu widget with the current known
        terminal size and the terminal.
        """
        self._window.draw(self._width, self._height, self.terminal)


    def _appButtons(self):
        for factory in self._store.powerupsFor(ITerminalServerFactory):
            yield Button(
                factory.name.encode('utf-8'),
                lambda factory=factory: self.switchTo(factory))


    def _logoffButton(self):
        return Button("logoff", self.logoff)


    def _makeWindow(self):
        buttons = VBox()
        for button in self._appButtons():
            buttons.addChild(Border(button))
        buttons.addChild(Border(self._logoffButton()))

        from twisted.internet import reactor
        window = TopWindow(self._draw, lambda f: reactor.callLater(0, f))
        window.addChild(Border(buttons))
        return window


    def connectionMade(self):
        """
        Reset the terminal and create a UI for selecting an application to use.
        """
        self.terminal.reset()
        self._window = self._makeWindow()


    def reactivate(self):
        """
        Called when a sub-protocol is finished.  This disconnects the
        sub-protocol and redraws the main menu UI.
        """
        self._protocol.connectionLost(None)
        self._protocol = None
        self.terminal.reset()
        self._window.filthy()
        self._window.repaint()


    def switchTo(self, app):
        """
        Use the given L{ITerminalServerFactory} to create a new
        L{ITerminalProtocol} and connect it to C{self.terminal} (such that it
        cannot actually disconnect, but can do most anything else).  Control of
        the terminal is delegated to it until it gives up that control by
        disconnecting itself from the terminal.

        @type app: L{ITerminalServerFactory} provider
        @param app: The factory which will be used to create a protocol
            instance.
        """
        viewer = _AuthenticatedShellViewer(list(getAccountNames(self._store)))
        self._protocol = app.buildTerminalProtocol(viewer)
        self._protocol.makeConnection(_ReturnToMenuWrapper(self, self.terminal))


    def keystrokeReceived(self, keyID, modifier):
        """
        Forward input events to the application-supplied protocol if one is
        currently active, otherwise forward them to the main menu UI.
        """
        if self._protocol is not None:
            self._protocol.keystrokeReceived(keyID, modifier)
        else:
            self._window.keystrokeReceived(keyID, modifier)


    def logoff(self):
        """
        Disconnect from the terminal completely.
        """
        self.terminal.loseConnection()



class _BetterTerminalSession(TerminalSession):
    """
    L{TerminalSession} is missing C{windowChanged} and C{eofReceived} for some
    reason.  Add it here until it's fixed in Twisted.  See Twisted ticket
    #3303.
    """
    def windowChanged(self, newWindowSize):
        """
        Ignore window size change events.
        """


    def eofReceived(self):
        """
        Ignore the eof event.
        """


class _BetterTerminalUser(TerminalUser):
    """
    L{TerminalUser} is missing C{conn} for some reason reason (probably the
    reason that it's not a very great thing and generally an implementation
    will be missing it for a while).  Add it here until it's fixed in Twisted.
    See Twisted ticket #3863.
    """
    # Some code in conch will rudely rebind this attribute later.  For now,
    # make sure that it is at least bound to something so that the object
    # appears to fully implement IConchUser.  Most likely, TerminalUser should
    # be taking care of this, not us.  Or even better, this attribute shouldn't
    # be part of the interface; some better means should be provided for
    # informing the IConchUser avatar of the connection object (I'm not even
    # sure why the avatar would care about having a reference to the connection
    # object).
    conn = None


class ShellAccount(Item):
    """
    Axiom cred hook for creating SSH avatars.
    """
    powerupInterfaces = (IConchUser,)
    implements(IPowerupIndirector)

    garbage = bytes()

    def indirect(self, interface):
        """
        Create an L{IConchUser} avatar which will use L{ShellServer} to
        interact with the connection.
        """
        if interface is IConchUser:
            componentized = Componentized()

            user = _BetterTerminalUser(componentized, None)
            session = _BetterTerminalSession(componentized)
            session.transportFactory = TerminalSessionTransport
            session.chainedProtocolFactory = lambda: ServerProtocol(ShellServer, self.store)

            componentized.setComponent(IConchUser, user)
            componentized.setComponent(ISession, session)

            return user

        raise NotImplementedError(interface)



class TerminalManhole(Item):
    """
    A terminal application which presents an interactive Python session running
    in the primary Mantissa server process.
    """
    powerupInterfaces = (ITerminalServerFactory,)
    implements(*powerupInterfaces)

    shell = dependsOn(ShellAccount)

    name = 'manhole'

    def buildTerminalProtocol(self, shellViewer):
        """
        Create and return a L{ColoredManhole} which includes this item's store
        in its namespace.
        """
        return ColoredManhole({'db': self.store, 'viewer': shellViewer})



__all__ = ['SecureShellConfiguration', 'ShellAccount', 'TerminalManhole']
